import re
import requests
from urllib.parse import urlparse

from difflib import SequenceMatcher

from coalib.results.Diff import Diff
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result
from coalib.bearlib import deprecate_settings
from coalib.settings.Setting import typed_list
from coalib.parsing.Globbing import fnmatch
from coalib.settings.Setting import typed_dict


class InvalidLinkBear(LocalBear):
    DEFAULT_TIMEOUT = 15
    LANGUAGES = {'All'}
    REQUIREMENTS = {PipRequirement('requests', '2.12')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Documentation'}

    # IP Address of www.google.com
    check_connection_url = 'http://216.58.218.174'

    @classmethod
    def check_prerequisites(cls):
        code = cls.get_status_code(
            cls.check_connection_url, cls.DEFAULT_TIMEOUT)
        return ('You are not connected to the internet.'
                if code is None else True)

    @staticmethod
    def get_status_code(url, timeout):
        try:
            code = requests.head(url, allow_redirects=False,
                                 timeout=timeout).status_code
            return code
        except requests.exceptions.RequestException:
            pass

    @staticmethod
    def parse_pip_vcs_url(link):
        splitted_at = link.split('@')[0]
        splitted_schema = splitted_at[splitted_at.index('+') + 1:]
        return splitted_schema

    @staticmethod
    def find_links_in_file(file, network_timeout, link_ignore_regex,
                           link_ignore_list):
        link_ignore_regex = re.compile(link_ignore_regex)
        regex = re.compile(
            r"""
            ((git\+|bzr\+|svn\+|hg\+|)  # For VCS URLs
            https?://                   # http:// or https:// as only these
                                        # are supported by the ``requests``
                                        # library
            [^.:%\s_/?#[\]@\\]+         # Initial part of domain
            \.                          # A required dot `.`
            (
                (?:[^\s()%\'"`<>|\\\[\]]+)  # Path name
                                            # This part does not allow
                                            # any parenthesis: balanced or
                                            # unbalanced.
            |                               # OR
                \([^\s()%\'"`<>|\\\[\]]*\)  # Path name contained within ()
                                        # This part allows path names that
                                        # are explicitly enclosed within one
                                        # set of parenthesis.
                                        # An example can be:
                                        # http://wik.org/Hello_(Adele_song)/200
            )
            *)
                                        # Thus, the whole part above
                                        # prevents matching of
                                        # Unbalanced parenthesis
            (?<!\.)(?<!,)               # Exclude trailing `.` or `,` from URL
            """, re.VERBOSE)
        for line_number, line in enumerate(file):
            match = regex.search(line)
            if match:
                link = match.group()
                if not (link_ignore_regex.search(link) or
                        fnmatch(link, link_ignore_list)):
                    if link.startswith(('hg+', 'bzr+', 'git+', 'svn+')):
                        link = InvalidLinkBear.parse_pip_vcs_url(link)
                    host = urlparse(link).netloc
                    code = InvalidLinkBear.get_status_code(
                               link,
                               network_timeout.get(host)
                               if host in network_timeout
                               else network_timeout.get('*')
                               if '*' in network_timeout
                               else InvalidLinkBear.DEFAULT_TIMEOUT)
                    yield line_number + 1, link, code

    @deprecate_settings(link_ignore_regex='ignore_regex',
                        network_timeout=('timeout', lambda t: {'*': t}))
    def run(self, filename, file,
            network_timeout: typed_dict(str, int, DEFAULT_TIMEOUT)=dict(),
            link_ignore_regex: str='([.\/]example\.com|\{|\$)',
            link_ignore_list: typed_list(str)='',
            follow_redirects: bool=False):
        """
        Find links in any text file and check if they are valid.

        A link is considered valid if the server responds with a 2xx code.

        This bear can automatically fix redirects, but ignores redirect
        URLs that have a huge difference with the original URL.

        Warning: This bear will make HEAD requests to all URLs mentioned in
        your codebase, which can potentially be destructive. As an example,
        this bear would naively just visit the URL from a line that goes like
        `do_not_ever_open = 'https://api.acme.inc/delete-all-data'` wiping out
        all your data.

        :param network_timeout:       A dict mapping URLs and timeout to be
                                      used for that URL. All the URLs that have
                                      the same host as that of URLs provided
                                      will be passed that timeout. It can also
                                      contain a wildcard timeout entry with key
                                      '*'. The timeout of all the websites not
                                      in the dict will be the value of the key
                                      '*'.
        :param link_ignore_regex:     A regex for urls to ignore.
        :param link_ignore_list: Comma separated url globs to ignore
        :param follow_redirects: Set to true to autocorrect redirects.
        """
        network_timeout = {urlparse(url).netloc
                           if not url == '*' else '*': timeout
                           for url, timeout in network_timeout.items()}

        for line_number, link, code in InvalidLinkBear.find_links_in_file(
                file, network_timeout, link_ignore_regex, link_ignore_list):
            if code is None:
                yield Result.from_values(
                    origin=self,
                    message=('Broken link - unable to connect to '
                             '{url}').format(url=link),
                    file=filename,
                    line=line_number,
                    severity=RESULT_SEVERITY.MAJOR)
            elif not 200 <= code < 300:
                # HTTP status 404, 410 or 50x
                if code in (404, 410) or 500 <= code < 600:
                    yield Result.from_values(
                        origin=self,
                        message=('Broken link - unable to connect to {url} '
                                 '(HTTP Error: {code})'
                                 ).format(url=link, code=code),
                        file=filename,
                        line=line_number,
                        severity=RESULT_SEVERITY.NORMAL)
                if follow_redirects and 300 <= code < 400:  # HTTP status 30x
                    redirect_url = requests.head(link,
                                                 allow_redirects=True).url
                    matcher = SequenceMatcher(
                        None, redirect_url, link)
                    if (matcher.real_quick_ratio() > 0.7 and
                            matcher.ratio()) > 0.7:
                        diff = Diff(file)
                        current_line = file[line_number - 1]
                        start = current_line.find(link)
                        end = start + len(link)
                        replacement = current_line[:start] + \
                            redirect_url + current_line[end:]
                        diff.change_line(line_number,
                                         current_line,
                                         replacement)

                        yield Result.from_values(
                            self,
                            'This link redirects to ' + redirect_url,
                            diffs={filename: diff},
                            file=filename,
                            line=line_number,
                            severity=RESULT_SEVERITY.NORMAL)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.settings.Setting import typed_list


@linter(executable='bootlint',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d*):(?P<column>\d*) (?P<severity>.)\d+ '
                     r'(?P<message>.+)')
class BootLintBear:
    """
    Raise several common HTML mistakes in html files that are using
    Bootstrap in a fairly "vanilla" way. Vanilla Bootstrap's components/widgets
    require their parts of the DOM to conform to certain structures that is
    checked. Also, raises issues about certain <meta> tags, HTML5 doctype
    declaration, etc. to use bootstrap properly.

    For more about the analysis, check Bootlint
    <https://github.com/twbs/bootlint#bootlint>.
    """
    LANGUAGES = {'HTML'}
    REQUIREMENTS = {NpmRequirement('bootlint', '0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         bootlint_ignore: typed_list(str)=[]):
        """
        :param bootlint_ignore: List of checkers to ignore.
        """
        ignore = ','.join(part.strip() for part in bootlint_ignore)
        return '--disable=' + ignore, filename



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.Result import Result
from coalib.settings.Setting import path


@linter(executable='tslint')
class TSLintBear:
    """
    Check TypeScript code for style violations and possible semantical
    problems.

    Read more about the capabilities at
    <https://github.com/palantir/tslint#core-rules>.
    """

    LANGUAGES = {'TypeScript'}
    REQUIREMENTS = {NpmRequirement('tslint', '3'),
                    NpmRequirement('typescript', '>=1.7.3')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/9re9c4fv17lhn7rmvzueebb3b'
    CAN_DETECT = {'Syntax', 'Formatting', 'Smell'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         tslint_config: path='', rules_dir: path=''):
        """
        :param tslint_config: Path to configuration file.
        :param rules_dir:     Rules directory
        """
        args = ('--format', 'json')
        if tslint_config:
            args += ('--config', tslint_config)
        if rules_dir:
            args += ('--rules-dir', rules_dir)
        return args + (filename,)

    def process_output(self, output, filename, file):
        output = json.loads(output) if output else []
        for issue in output:
            yield Result.from_values(
                origin='{} ({})'.format(self.__class__.__name__,
                                        issue['ruleName']),
                message=issue['failure'],
                file=issue['name'],
                line=int(issue['startPosition']['line']) + 1,
                end_line=int(issue['endPosition']['line']) + 1,
                column=int(issue['startPosition']['character']) + 1,
                end_column=int(issue['endPosition']['character']) + 1)



from coalib.bearlib.abstractions.Linter import linter
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='phpcs',
        output_format='regex',
        config_suffix='.xml',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): '
                     r'\w+ - (?P<message>.+)')
class PHPCodeSnifferBear:
    """
    Ensures that your PHP, JavaScript or CSS code remains clean and consistent.

    See <https://github.com/squizlabs/PHP_CodeSniffer> for more information.
    """

    LANGUAGES = {'PHP', 'JavaScript', 'CSS'}
    REQUIREMENTS = {DistributionRequirement(apt_get='php-codesniffer')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax', 'Documentation',
                  'Code Simplification'}
    ASCIINEMA_URL = 'https://asciinema.org/a/efawv96vdalck73tc3hwcabov'

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--report=emacs', '--standard='+config_file, filename

    @staticmethod
    def generate_config(filename, file,
                        max_line_length: int=79,
                        line_ending_character: str='\\n',
                        indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
                        use_spaces: bool=True,
                        allow_multiple_statements_per_line: bool=False,
                        force_lower_case_keywords: bool=True,
                        force_lower_case_constants: bool=True,
                        blank_line_after_namespace_declaration: bool=True,
                        check_use_blocks: bool=True,
                        check_class_declaration: bool=True,
                        check_property_declaration: bool=True,
                        force_scope_modifier_on_method: bool=True,
                        function_declaration_argument_spacing: int=1,
                        allow_multiline_function_declaration: bool=True):
        """
        :param max_line_length:
            Maximum number of characters for a line.
        :param line_ending_character:
            Checks that end of line characters correspond to the one provided.
        :param indent_size:
            Number of spaces per indentation level.
        :param use_spaces:
            True if spaces are to be used instead of tabs.
        :param allow_multiple_statements_per_line:
            Allows having multiple statements on one line.
        :param force_lower_case_keyword:
            Checks that ``PHP`` keywords are lowercase.
        :param force_lower_case_constant:
            Checks that all uses of ``true``, ``false`` and ``null`` are
            lowercase.
        :param blank_line_after_namespace_declaration:
            Ensures that there is a blank line after a namespace declaration.
        :param check_use_blocks:
            Ensures that there is one blank line after a ``use`` block,
            that there is only one use block per line, and that all ``use``
            declaration are done after namespaces declaration.
        :param check_class_declaration:
            Ensures that ``extends`` and ``implements`` keywords are declared
            on the same line as the class name, that the opening brace for a
            class is on the next line, and that the closing brace for a class
            is on the next line after the body. Allows splitting implements
            list accross multiple lines.
        :param check_property_declaration:
            Ensures that visibility is declared on all properties, that the
            ``var`` keyword is not used to declare a property, that there is
            not more that one property declared on a line, that properties are
            not prefixed with an underscore.
        :param force_scope_modifier_on_method:
            Verifies that class methods have scope modifiers.
        :param function_declaration_argument_spacing:
            Number of spaces between arguments in function declaration.
        :param allow_multiline_function_declaration:
            Allows argument lists to be split accross multiple lines correctly
            indented.
        """
        rules_map = {'Generic.WhiteSpace.DisallowTabIndent':
                     use_spaces,
                     'Generic.Formatting.DisallowMultipleStatements':
                     not allow_multiple_statements_per_line,
                     'Generic.PHP.LowerCaseKeyword':
                     force_lower_case_keywords,
                     'Generic.PHP.LowerCaseConstant':
                     force_lower_case_constants,
                     'PSR2.Namespaces.UseDeclaration':
                     check_use_blocks,
                     'PSR2.Namespaces.NamespaceDeclaration':
                     blank_line_after_namespace_declaration,
                     'PSR2.Classes.ClassDeclaration':
                     check_class_declaration,
                     'PSR2.Classes.PropertyDeclaration':
                     check_property_declaration,
                     'Squiz.Scope.MethodScope':
                     force_scope_modifier_on_method,
                     'Squiz.Functions.MultiLineFunctionDeclaration':
                     allow_multiline_function_declaration}
        rules = ''
        for k, v in rules_map.items():
            rules += '<rule ref="{}"/>\n'.format(k) if v else ''
        configs = '''<?xml version="1.0"?>
<ruleset name="Custom Standard">
 <description>A custom coding standard</description>
 <!-- Include the whole PSR-1 standard -->
 <rule ref="PSR1"/>
 <rule ref="Generic.Files.LineLength">
  <properties>
   <property name="lineLimit" value="{max_line_length}"/>
  </properties>
 </rule>
 <rule ref="Generic.Files.LineEndings">
  <properties>
   <property name="eolChar" value="{line_ending_character}"/>
  </properties>
 </rule>
 <rule ref="Generic.WhiteSpace.ScopeIndent">
  <properties>
   <property name="ignoreIndentationTokens" type="array"
   value="T_COMMENT,T_DOC_COMMENT_OPEN_TAG"/>
  </properties>
  <properties>
   <property name="indent" value="{indent_size}"/>
  </properties>
  <properties>
   <property name="exact" value="true"/>
  </properties>
 </rule>
 {some_rules}
 <rule ref="Squiz.WhiteSpace.ScopeKeywordSpacing"/>
 <rule ref="Squiz.Functions.FunctionDeclarationArgumentSpacing">
  <properties>
   <property name="equalsSpacing"
   value="{function_declaration_argument_spacing}"/>
  </properties>
 </rule>
 <rule
 ref="Squiz.Functions.FunctionDeclarationArgumentSpacing.SpacingAfterHint">
  <severity>0</severity>
 </rule>
</ruleset>
'''.format(max_line_length=max_line_length,
           line_ending_character=line_ending_character,
           indent_size=indent_size,
           some_rules=rules,
           function_declaration_argument_spacing=(
               function_declaration_argument_spacing))
        return configs



import re

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='csvlint')
class CSVLintBear:
    """
    Verifies using ``csvlint`` if ``.csv`` files are valid CSV or not.
    """

    LANGUAGES = {'CSV'}
    REQUIREMENTS = {GemRequirement('csvlint')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}
    ASCIINEMA_URL = 'https://asciinema.org/a/8fmp2pny34kpqw7t1eoy7phhc'

    regex = re.compile(r'\n\d+\.\s(?P<origin>(?P<severity>\w+))\.\s'
                       r'(Row:\s(?P<line>[0-9]+)\.\s)?(?P<message>.*)?')

    severity_map = {
        'wrong_content_type': RESULT_SEVERITY.MAJOR,
        'ragged_rows': RESULT_SEVERITY.MAJOR,
        'blank_rows': RESULT_SEVERITY.MAJOR,
        'invalid_encoding': RESULT_SEVERITY.MAJOR,
        'not_found': RESULT_SEVERITY.MAJOR,
        'stray_quote': RESULT_SEVERITY.MAJOR,
        'unclosed_quote': RESULT_SEVERITY.MAJOR,
        'whitespace': RESULT_SEVERITY.MAJOR,
        'line_breaks': RESULT_SEVERITY.MAJOR,
        'no_encoding': RESULT_SEVERITY.NORMAL,
        'encoding': RESULT_SEVERITY.NORMAL,
        'no_content_type': RESULT_SEVERITY.NORMAL,
        'excel': RESULT_SEVERITY.NORMAL,
        'check_options': RESULT_SEVERITY.NORMAL,
        'inconsistent_values': RESULT_SEVERITY.NORMAL,
        'empty_column_name': RESULT_SEVERITY.NORMAL,
        'duplicate_column_name': RESULT_SEVERITY.NORMAL,
        'title_row': RESULT_SEVERITY.NORMAL,
        'nonrfc_line_breaks': RESULT_SEVERITY.INFO,
        'assumed_header': RESULT_SEVERITY.INFO}

    message_dict = {
        'wrong_content_type': 'Content type is not text/csv.',
        'ragged_rows': 'Row has a different number of columns. (than the first'
                       ' row in the file)',
        'blank_rows': 'Completely empty row, e.g. blank line or a line where'
                      ' all column values are empty.',
        'invalid_encoding': 'Encoding error when parsing row, e.g. because of'
                            ' invalid characters.',
        'not_found': 'HTTP 404 error when retrieving the data.',
        'stray_quotd': 'Missing or stray quote.',
        'unclosed_quotd': 'Unclosed quoted field.',
        'whitespacd': 'A quoted column has leading or trailing whitespace.',
        'line_breakd': 'Line breaks were inconsistent or incorrectly'
                       ' specified.',
        'no_encodind': 'The Content-Type header returned in the HTTP request'
                       ' does not have a charset parameter.',
        'encoding': 'The character set is not UTF-8.',
        'no_content_type': 'File is being served without a Content-Type'
                           ' header.',
        'excel': 'No Content-Type header and the file extension is .xls.',
        'check_optiond': 'CSV file appears to contain only a single column.',
        'inconsistent_valued': 'Inconsistent values in the same column.'
                               ' Reported if <90% of values seem to have same'
                               ' data type. (either numeric or alphanumeric'
                               ' including punctuation)',
        'empty_column_name': 'A column in the CSV header has an empty name.',
        'duplicate_column_name': 'A column in the CSV header has a duplicate'
                                 ' name.',
        'title_rod': 'There appears to be a title field in the first row of'
                      ' the CSV.',
        'nonrfc_line_breakd': 'Uses non-CRLF line breaks, so does not conform'
                              ' to RFC4180.',
        'assumed_headed': 'The validator has assumed that a header is present.'
    }

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,

    @classmethod
    def process_output(self, output, filename, file, result_message=None):
        for match in re.finditer(self.regex, str(output)):
            groups = match.groupdict()
            result_message = ' ' + groups['message'] if groups[
                'line'] is None else ''
            yield self._convert_output_regex_match_to_result(
                self,
                match, filename, severity_map=self.severity_map,
                result_message=self.message_dict[groups['origin']] +
                result_message)



from coalib.bears.LocalBear import LocalBear
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class LineCountBear(LocalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    def run(self, filename, file, max_lines_per_file: int):
        """
        Count the number of lines in a file and ensure that they are
        smaller than a given size.

        :param max_lines_per_file: Number of lines allowed per file.
        """
        file_length = len(file)
        if file_length > max_lines_per_file:
            yield Result.from_values(
                origin=self,
                message=('This file had {count} lines, which is {extra} '
                         'lines more than the maximum limit specified.'
                         .format(count=file_length,
                                 extra=file_length-max_lines_per_file)),
                severity=RESULT_SEVERITY.NORMAL,
                file=filename)



from collections import namedtuple
import textwrap
import sys

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement


class FlagInfo(namedtuple('FlagInfo', 'arg doc inverse')):
    """
    Information about a command-line flag.

    :param arg:
        The argument to pass to enable the flag.
    :param doc:
        A help text for the flag.
    :param inverse:
        Set to ``True`` when the coala option is the inverse of
        the subprocess one, for example coala's ``allow_untyped_calls``
        is the inverse of mypy's ``--disallow-untyped-calls``.
    """

    def want_flag(self, value):
        """
        Check if the flag should be added to the argument list.

        :param value: The configuration value.
        :return:      The flag value, may be negated if the flag specifies so.
        """

        if self.inverse:
            value = not value
        return value


FLAG_MAP = {
    'allow_untyped_functions': FlagInfo(
        arg='--disallow-untyped-defs',
        doc='Allow defining functions without type annotations or with '
            'incomplete type annotations.',
        inverse=True),
    'allow_untyped_calls': FlagInfo(
        arg='--disallow-untyped-calls',
        doc='Allow calling functions without type annotations from '
            'typed functions.',
        inverse=True),
    'check_untyped_function_bodies': FlagInfo(
        arg='--check-untyped-defs',
        doc='Do not check the interior of functions without type annotations.',
        inverse=False),
    'strict_optional': FlagInfo(
        arg='--strict-optional',
        doc='Enable experimental strict checks related to Optional types. See '
            '<http://mypy-lang.blogspot.com.es/2016/07/mypy-043-released.html>'
            ' for an explanation.',
        inverse=False),
}


def add_param_docs(param_map):
    """
    Append documentation from FLAG_MAP to a function's docstring.

    :param param_map:
        A mapping of argument names (strings) to FlagInfo objects.
    :return:
        A decorator that appends flag information to a function's docstring.
    """
    def decorator(func):
        func.__doc__ = textwrap.dedent(func.__doc__) + '\n'.join(
            ':param {}:\n{}'.format(name, textwrap.indent(arg.doc, '    '))
            for name, arg in param_map.items())
        return func
    return decorator


# Mypy generates messages in the format:
#    blabla.py: note: In function "f":
#    blabla.py:2: error: Unsupported operand types for ...
# The "note" messages are only adding info coala should already know,
# so discard those. We're only capturing the errors.
@linter(executable=sys.executable,
        prerequisite_check_command=(sys.executable, '-m', 'mypy', '-V'),
        output_format='regex',
        output_regex=r'[^:]+:(?:(?P<line>\d+):)? '
                      '(?P<severity>error): (?P<message>.*)')
class MypyBear:
    """
    Type-checks your Python files!

    Checks optional static typing using the mypy tool.
    See <http://mypy.readthedocs.io/en/latest/basics.html> for info on how to
    add static typing.
    """

    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    AUTHORS = {'Petr Viktorin'}
    REQUIREMENTS = {PipRequirement('mypy-lang', '0.4.6')}
    AUTHORS_EMAILS = {'encukou@gmail.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/90736'

    # This detects typing errors, which is pretty unique -- it doesn't
    # make sense to add a category for it.
    CAN_DETECT = set()

    @add_param_docs(FLAG_MAP)
    def create_arguments(self, filename, file, config_file,
                         language: str='Python 3',
                         python_version: str=None,
                         allow_untyped_functions: bool=True,
                         allow_untyped_calls: bool=True,
                         check_untyped_function_bodies: bool=False,
                         strict_optional: bool=False):
        """
        :param language:
            Set to ``Python`` or ``Python 3`` to check Python 3.x source.
            Use ``Python 2`` for Python 2.x.
        :param python_version:
            Set the specific Python version, e.g. ``3.5``.
        """
        args = ['-m', 'mypy']
        if language.lower() == 'python 2':
            args.append('--py2')
        elif language.lower() not in ('python 3', 'python'):
            # Ideally, this would fail the check, but there's no good
            # way to fail from create_arguments.
            # See https://github.com/coala-analyzer/coala/issues/2573
            self.err(
                'Language needs to be "Python", "Python 2" or "Python 3". '
                'Assuming Python 3.')
        if python_version:
            args.extend(['--python-version', python_version])
        loc = locals()
        args.extend(flag.arg for name, flag in FLAG_MAP.items()
                    if flag.want_flag(loc[name]))
        args.append(filename)
        return args



import json
import re

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.misc.Compatibility import JSONDecodeError
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='cr')
class JSComplexityBear:
    """
    Calculates cyclomatic complexity using ``cr``, the command line utility
    provided by the NodeJS module ``complexity-report``.
    """

    LANGUAGES = {'JavaScript'}
    REQUIREMENTS = {NpmRequirement('complexity-report', '2.0.0-alpha')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/39250'
    CAN_DETECT = {'Complexity'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--format', 'json', filename

    def process_output(self, output, filename, file, cc_threshold: int=10):
        """
        :param cc_threshold: Threshold value for cyclomatic complexity
        """
        message = '{} has a cyclomatic complexity of {}.'
        if output:
            try:
                output = json.loads(output)
            except JSONDecodeError:
                output_regex = (r'Fatal error \[getReports\]: .+: '
                                r'Line (?P<line>\d+): (?P<message>.*)')
                for match in re.finditer(output_regex, output):
                    groups = match.groupdict()
                    yield Result.from_values(
                        origin=self,
                        message=groups['message'].strip(),
                        file=filename,
                        severity=RESULT_SEVERITY.MAJOR,
                        line=int(groups['line']))
                return
            for function in output['reports'][0]['functions']:
                if function['cyclomatic'] >= cc_threshold:
                    yield Result.from_values(
                        origin=self,
                        message=message.format(function['name'],
                                               function['cyclomatic']),
                        file=filename,
                        line=function['line'])



import json
import yaml

from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='rubocop',
        use_stdin=True)
class RuboCopBear:
    """
    Check Ruby code for syntactic, formatting as well as semantic problems.

    See <https://github.com/bbatsov/rubocop#cops> for more information.
    """

    LANGUAGES = {'Ruby'}
    REQUIREMENTS = {GemRequirement('rubocop'),
                    PipRequirement('pyyaml', '3.12')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/39241'
    CAN_DETECT = {'Simplification'}
    CAN_FIX = {'Syntax', 'Formatting'}

    severity_map = {'error': RESULT_SEVERITY.MAJOR,
                    'warning': RESULT_SEVERITY.NORMAL,
                    'convention': RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file, rubocop_config: str=''):
        # Need both stdin and filename. Explained in this comment:
        # https://github.com/bbatsov/rubocop/pull/2146#issuecomment-131403694
        args = (filename, '--stdin', '--format=json')
        if rubocop_config:
            args += ('--config', rubocop_config)
        else:
            args += ('--config', config_file)
        return args

    @staticmethod
    @deprecate_settings(indent_size='tab_width',
                        method_length_count_comments='method_count_comments',
                        method_naming_convention='method_name_case',
                        variable_naming_convention='variable_name_case')
    def generate_config(filename, file,
                        access_modifier_indentation: str='indent',
                        preferred_alias: str='prefer_alias',
                        align_hash_rocket_by: str='key',
                        align_colon_by: str='key',
                        inspect_last_argument_hash: str='always_inspect',
                        align_parameters: str='with_first_parameter',
                        class_check: str='is_a?',
                        comment_keywords: tuple=('TODO',
                                                 'FIXME',
                                                 'OPTIMIZE',
                                                 'HACK',
                                                 'REVIEW'),
                        min_if_unless_guard: int=1,
                        indent_size: int=2,
                        method_naming_convention: str='snake',
                        string_literals: str='single_quotes',
                        variable_naming_convention: str='snake',
                        max_class_length: int=100,
                        class_length_count_comments: bool=False,
                        max_module_length: int=100,
                        module_length_count_comments: bool=False,
                        cyclomatic_complexity: int=6,
                        max_line_length: int=79,
                        line_length_allow_here_doc: bool=True,
                        line_length_allow_uri: bool=True,
                        max_method_length: int=10,
                        method_length_count_comments: bool=False,
                        max_parameters: int=5,
                        count_keyword_args: bool=True,
                        ignore_unused_block_args_if_empty: bool=True,
                        allow_unused_block_keyword_arguments: bool=False,
                        ignore_unused_method_args_if_empty: bool=True,
                        allow_unused_method_keyword_args: bool=False):
        """
        Not all settings added.
        Notable settings missing: Rails settings.

        :param access_modifier_indentation:
            Indent private/protected/public as deep as method definitions
            options:
                ``indent`` :  Indent modifiers like class members.
                ``outdent`` : Indent modifiers one level less than
                              class members.
        :param preferred_alias:
            Which method to use for aliasing in ruby.
            options : ``alias`` , ``alias_method``.
        :param align_hash_rocket_by:
            Alignment of entries using hash rocket as separator.
        :param align_colon_by:
            Alignment of entries using colon as separator.
        :param inspect_last_argument_hash:
            Select whether hashes that are the last argument in a method call
            should be inspected.
            options: ``always_inspect``, ``always_ignore``,
                     ``ignore_implicit``, ``ignore_explicit``.
        :param align_parameters:
            Alignment of parameters in multi-line method calls.

            options:
                ``with_first_parameter``: Aligns the following lines
                                          along the same column as the
                                          first parameter.

                ``with_fixed_indentation``: Aligns the following lines with one
                                            level of indentation relative to
                                            the start of the line with the
                                            method call.
        :param class_check:
            How to check type of class.
            options: ``is_a?``, ``kind_of?``.
        :param comment_keywords:
            Checks formatting of special comments based on keywords like
            TODO, FIXME etc.
        :param min_if_unless_guard:
            The number of lines that are tolerable within an if/unless block,
            more than these lines call for the usage of a guard clause.
        :param indent_size:
            Number of spaces per indentation level.
        :param method_naming_convention:
            Case of a method's name.
            options: ``snake``, ``camel``.
        :param string_literals:
            Use ' or " as string literals.
            options: ``single_quotes``, ``double_quotes``.
        :param variable_naming_convention:
            Case of a variable's name.
            options: ``snake``, ``camel``.
        :param max_class_length:
            Max lines in a class.
        :param class_length_count_comments:
            Whether or not to count comments while calculating the class
            length.
        :param max_module_length:
            Max lines in a module.
        :param module_length_count_comments:
            Whether or not to count comments while calculating
            the module length.
        :param cyclomatic_complexity:
            Cyclomatic Complexity of the file.
        :param max_line_length:
            Max length of a line.
        :param line_length_allow_here_doc:
            Allow here-doc lines to be more than the max line length.
        :param line_length_allow_uri:
            To make it possible to copy or click on URIs in the code,
            we allow ignore long lines containing a URI to be longer than max
            line length.
        :param max_method_length:
            Max number of lines in a method.
        :param method_length_count_comments:
            Whether or not to count full line comments while calculating
            method length.
        :param max_parameters:
            Max number of parameters in parameter list.
        :param count_keyword_args:
            Count keyword args while counting all arguments?
        :param ignore_unused_block_args_if_empty:
            Ignore unused block arguments if block is empty.
        :param allow_unused_block_keyword_arguments:
            Allow unused block keyword arguments.
        :param ignore_unused_method_args_if_empty:
            Allows unused method argument if method is empty.
        :param allow_unused_method_keyword_args:
            Allows unused keyword arguments in a method.
        """
        naming_convention = {'camel': 'camelCase', 'snake': 'snake_case'}
        options = {
            'Style/AccessModifierIndentation': {
                'EnforcedStyle': access_modifier_indentation
            },
            'Style/Alias': {
                'EnforcedStyle': preferred_alias
            },
            'Style/AlignHash': {
                'EnforcedHashRocketStyle': align_hash_rocket_by,
                'EnforcedColonStyle': align_colon_by,
                'EnforcedLastArgumentHashStyle': inspect_last_argument_hash
            },
            'Style/AlignParameters': {
                'EnforcedStyle': align_parameters
            },
            'Style/ClassCheck': {
                'EnforcedStyle': class_check
            },
            'Style/CommentAnnotation': {
                'Keywords': comment_keywords
            },
            'Style/GuardClause': {
                'MinBodyLength': min_if_unless_guard
            },
            'Style/IndentationWidth': {
                'Width': indent_size
            },
            'Style/MethodName': {
                'EnforcedStyle': naming_convention.get(
                                 method_naming_convention,
                                 method_naming_convention)
            },
            'Style/StringLiterals': {
                 'EnforcedStyle': string_literals
            },
            'Style/VariableName': {
                'EnforcedStyle': naming_convention.get(
                                 variable_naming_convention,
                                 variable_naming_convention)
            },
            'Metrics/ClassLength': {
                'Max': max_class_length,
                'CountComments': class_length_count_comments
            },
            'Metrics/ModuleLength': {
                'CountComments': module_length_count_comments,
                'Max': max_module_length
            },
            'Metrics/CyclomaticComplexity': {
                'Max': cyclomatic_complexity
            },
            'Metrics/LineLength': {
                'Max': max_line_length,
                'AllowHeredoc': line_length_allow_here_doc,
                'AllowURI': line_length_allow_uri
            },
            'Metrics/MethodLength': {
                'CountComments':  method_length_count_comments,
                'Max': max_method_length
            },
            'Metrics/ParameterLists': {
                'Max': max_parameters,
                'CountKeywordArgs': count_keyword_args
            },
            'Lint/UnusedBlockArgument': {
                'IgnoreEmptyBlocks': ignore_unused_block_args_if_empty,
                'AllowUnusedKeywordArguments':
                    allow_unused_block_keyword_arguments
            },
            'Lint/UnusedMethodArgument': {
                'AllowUnusedKeywordArguments':
                    allow_unused_method_keyword_args,
                'IgnoreEmptyMethods':
                    ignore_unused_method_args_if_empty
            },
        }
        return yaml.dump(options, default_flow_style=False)

    def process_output(self, output, filename, file):
        output = json.loads(output)
        assert len(output['files']) == 1
        for result in output['files'][0]['offenses']:
            # TODO: Add condition for auto-correct, when rubocop is updated.
            # Relevant Issue: https://github.com/bbatsov/rubocop/issues/2932
            yield Result.from_values(
                origin='{class_name} ({rule})'.format(
                    class_name=self.__class__.__name__,
                    rule=result['cop_name']),
                message=result['message'],
                file=filename,
                diffs=None,
                severity=self.severity_map[result['severity']],
                line=result['location']['line'],
                column=result['location']['column'],
                # Tested with linebreaks, it's secure.
                end_column=result['location']['column'] +
                result['location']['length'])



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement


@linter(executable='sqlint', use_stdin=True, output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+):'
                     r'(?P<severity>ERROR|WARNING) (?P<message>(?:\s*.+)*)')
class SQLintBear:
    """
    Check the given SQL files for syntax errors or warnings.

    This bear supports ANSI syntax. Check out
    <https://github.com/purcell/sqlint> for more detailed information.
    """

    LANGUAGES = {'SQL'}
    REQUIREMENTS = {GemRequirement('sqlint')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ()



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='mcs',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.+\((?P<line>\d+),(?P<column>\d+)\): '
                     r'(?P<severity>error|warning) \w+: (?P<message>.+)')
class CSharpLintBear:
    """
    Checks C# code for syntactical correctness using the ``mcs`` compiler.
    """

    LANGUAGES = {'C#'}
    REQUIREMENTS = {DistributionRequirement(apt_get='mono-mcs')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



from restructuredtext_lint import lint

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class reSTLintBear(LocalBear):
    LANGUAGES = {'reStructuredText'}
    REQUIREMENTS = {PipRequirement('restructuredtext-lint', '0.17.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting', 'Syntax'}

    def run(self, filename, file):
        """
        Lints reStructuredText.
        """
        content = ''.join(file)
        errors = lint(content)

        for error in errors:
            severity = {
                1: RESULT_SEVERITY.INFO,
                2: RESULT_SEVERITY.NORMAL,
                3: RESULT_SEVERITY.MAJOR,
                4: RESULT_SEVERITY.MAJOR}.get(error.level,
                                              RESULT_SEVERITY.NORMAL)
            yield Result.from_values(
                self,
                error.message,
                file=filename,
                line=error.line,
                debug_msg=error.full_message,
                severity=severity)



import autopep8
import nbformat

from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.settings.Setting import typed_list

# Comments regarind Jupyter Notebooks:
# The `nbformat` module contains the reference implementation of the Jupyter
# Notebook format, and Python APIs for working with notebooks.
# On the file level, a notebook is a JSON file, i.e. dictionary with a few
# keys.
# The functions in `nbformat` work with `NotebookNode` objects, which are like
# dictionaries, but allow attribute access. The structure of these objects
# matches the notebook format specification.


def notebook_node_from_string_list(string_list):
    """
    Reads a notebook from a string list and returns the NotebookNode
    object.

    :param string_list: The notebook file contents as list of strings
                        (linewise).
    :return:            The notebook as NotebookNode.
    """
    return nbformat.reads(''.join(string_list), nbformat.NO_CONVERT)


def notebook_node_to_string_list(notebook_node):
    """
    Writes a NotebookNode to a list of strings.

    :param notebook_node: The notebook as NotebookNode to write.
    :return:              The notebook as list of strings (linewise).
    """
    return nbformat.writes(notebook_node, nbformat.NO_CONVERT).splitlines(True)


def autopep8_fix_code_cell(source, options=None, apply_config=None):
    """
    Applies autopep8.fix_code and takes care of newline characters.

    autopep8.fix_code automatically adds a final newline at the end,
    e.g. ``autopep8.fix_code('a=1')`` yields 'a = 1\\n'.
    Note that this is not related to the 'W292' flag, i.e.
    ``autopep8.fix_code('a=1', options=dict(ignore=('W292',)))`` gives
    the same result.
    For notebook code cells, this behaviour does not make sense, hence
    newline is removed if ``source`` does not end with one.
    """
    source_corrected = autopep8.fix_code(source,
                                         apply_config=apply_config,
                                         options=options)

    if not source.endswith('\n'):
        return source_corrected[:-1]

    return source_corrected


class PEP8NotebookBear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('autopep8', '1.2'),
                    PipRequirement('nbformat', '4.1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/83333'
    CAN_FIX = {'Formatting'}

    def run(self, filename, file,
            max_line_length: int=79,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            pep_ignore: typed_list(str)=(),
            pep_select: typed_list(str)=(),
            local_pep8_config: bool=False):
        """
        Detects and fixes PEP8 incompliant code in Jupyter Notebooks. This bear
        will not change functionality of the code in any way.

        :param max_line_length:   Maximum number of characters for a line.
        :param indent_size:       Number of spaces per indent level.
        :param pep_ignore:        A list of errors/warnings to ignore.
        :param pep_select:        A list of errors/warnings to exclusively
                                  apply.
        :param local_pep8_config: Set to true if autopep8 should use a config
                                  file as if run normally from this directory.
        """
        options = {'ignore': pep_ignore,
                   'select': pep_select,
                   'max_line_length': max_line_length,
                   'indent_size': indent_size}
        notebook_node = notebook_node_from_string_list(file)
        cells = notebook_node['cells']

        for cell in cells:
            if cell['cell_type'] != 'code':
                continue
            cell['source'] = autopep8_fix_code_cell(cell['source'],
                                                    local_pep8_config,
                                                    options)

        corrected = notebook_node_to_string_list(notebook_node)

        # If newline at eof in `file` but not in `corrected`, add
        # final newline character to `corrected` to make sure this difference
        # does not pop up in `diffs`.
        if file[-1].endswith('\n') and not corrected[-1].endswith('\n'):
            corrected[-1] += '\n'

        diffs = Diff.from_string_arrays(file, corrected).split_diff()

        for diff in diffs:
            yield Result(self,
                         'The code does not comply to PEP8.',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff})



from coalib.bearlib.languages.LanguageDefinition import LanguageDefinition
from coalib.bears.LocalBear import LocalBear
from coalib.results.HiddenResult import HiddenResult
from coalib.results.Result import Result, RESULT_SEVERITY
from coalib.results.SourceRange import SourceRange
from coalib.results.AbsolutePosition import AbsolutePosition
from coala_utils.string_processing.Core import unescaped_search_for


class AnnotationBear(LocalBear):
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'

    def run(self, filename, file, language: str, coalang_dir: str = None):
        """
        Finds out all the positions of strings and comments in a file.
        The Bear searches for valid comments and strings and yields their
        ranges as SourceRange objects in HiddenResults.

        :param language:
            The programming language of the source code.
        :param coalang_dir:
            External directory for coalang file.
        :return:
            One HiddenResult containing a dictionary with keys being 'strings'
            or 'comments' and values being a tuple of SourceRanges pointing to
            the strings and a tuple of SourceRanges pointing to all comments
            respectively. The ranges do include string quotes or the comment
            starting separator but not anything before (e.g. when using
            ``u"string"``, the ``u`` will not be in the source range).
        """
        try:
            lang_dict = LanguageDefinition(language, coalang_dir=coalang_dir)
        except FileNotFoundError:
            content = ('coalang specification for ' + language +
                       ' not found.')
            yield HiddenResult(self, content)
            return

        string_delimiters = dict(lang_dict['string_delimiters'])
        multiline_string_delimiters = dict(
            lang_dict['multiline_string_delimiters'])
        multiline_comment_delimiters = dict(
            lang_dict['multiline_comment_delimiters'])
        comment_delimiter = dict(lang_dict['comment_delimiter'])
        string_ranges = comment_ranges = ()
        try:
            string_ranges, comment_ranges = self.find_annotation_ranges(
                file,
                filename,
                string_delimiters,
                multiline_string_delimiters,
                comment_delimiter,
                multiline_comment_delimiters)

        except NoCloseError as e:
            yield Result(self, str(e), severity=RESULT_SEVERITY.MAJOR,
                         affected_code=(e.code,))

        content = {'strings': string_ranges, 'comments': comment_ranges}
        yield HiddenResult(self, content)

    def find_annotation_ranges(self,
                               file,
                               filename,
                               string_delimiters,
                               multiline_string_delimiters,
                               comment_delimiter,
                               multiline_comment_delimiters):
        """
        Finds ranges of all annotations.

        :param file:
            A tuple of strings, with each string being a line in the file.
        :param filename:
            The name of the file.
        :param string_delimiters:
            A dictionary containing the various ways to  define single-line
            strings in a language.
        :param multiline_string_delimiters:
            A dictionary containing the various ways to define multi-line
            strings in a language.
        :param comment_delimiter:
            A dictionary containing the various ways to define single-line
            comments in a language.
        :param multiline_comment_delimiters:
            A dictionary containing the various ways to define multi-line
            comments in a language.
        :return:
            Two tuples first containing a tuple of strings, the second a tuple
            of comments.
        """
        text = ''.join(file)
        strings_range = []
        comments_range = []
        position = 0
        while position <= len(text):

            def get_new_position():
                _range, end_position = self.get_range_end_position(
                    file,
                    filename,
                    text,
                    multiline_string_delimiters,
                    position,
                    self.get_multiline)
                if end_position and _range:
                    strings_range.append(_range)
                    return end_position + 1

                _range, end_position = self.get_range_end_position(
                    file,
                    filename,
                    text,
                    string_delimiters,
                    position,
                    self.get_singleline_strings)
                if end_position and _range:
                    strings_range.append(_range)
                    return end_position + 1

                _range, end_position = self.get_range_end_position(
                    file,
                    filename,
                    text,
                    multiline_comment_delimiters,
                    position,
                    self.get_multiline)
                if end_position and _range:
                    comments_range.append(_range)
                    return end_position + 1

                _range, end_position = self.get_range_end_position(
                    file,
                    filename,
                    text,
                    comment_delimiter,
                    position,
                    self.get_singleline_comment,
                    single_comment=True)
                if end_position and _range:
                    comments_range.append(_range)
                    return end_position + 1

                return position + 1

            position = get_new_position()

        return tuple(strings_range), tuple(comments_range)

    @staticmethod
    def get_range_end_position(file,
                               filename,
                               text,
                               annotations,
                               position,
                               func,
                               single_comment=False):
        _range = end_position = None
        for annotation in annotations.keys():
            if text[position:].startswith(annotation):
                if not single_comment:
                    ret_val = func(file,
                                   filename,
                                   text,
                                   annotation,
                                   annotations[annotation],
                                   position)
                else:
                    ret_val = func(file,
                                   filename,
                                   text,
                                   annotation,
                                   position)
                if ret_val:
                    _range, end_position = ret_val[0], ret_val[1]

        return _range, end_position

    @staticmethod
    def get_multiline(file,
                      filename,
                      text,
                      annotation_start,
                      annotation_end,
                      position):
        """
        Gets sourcerange and end position of an annotation that can span
        multiple lines.

        :param file:
            A tuple of strings, with each string being a line in the file.
        :param filename:
            The name of the file.
        :param annotation_start:
            The string specifying the start of the annotation.
        :param annotation_end:
            The string specifying the end of the annotation.
        :param position:
            An integer identifying the position where the annotation started.
        :return:
            A SourceRange object holding the range of the multi-line annotation
            and the end_position of the annotation as an integer.
        """
        end_end = get_end_position(annotation_end,
                                   text,
                                   position + len(annotation_start) - 1)
        if end_end == -1:
            _range = SourceRange.from_absolute_position(
                filename,
                AbsolutePosition(file, position))
            raise NoCloseError(annotation_start, _range)

        return (SourceRange.from_absolute_position(
                    filename,
                    AbsolutePosition(file, position),
                    AbsolutePosition(file, end_end)),
                end_end)

    @staticmethod
    def get_singleline_strings(file,
                               filename,
                               text,
                               string_start,
                               string_end,
                               position):
        """
        Gets sourcerange of a single-line string and its end position.

        :param file:
            A tuple of strings, with each string being a line in the file.
        :param filename:
            The name of the file.
        :param string_start:
            The string which specifies how a string starts.
        :param string_end:
            The string which specifies how a string ends.
        :position:
            An integer identifying the position where the string started.
        :return:
            A SourceRange object identifying the range of the single-line
            string and the end_position of the string as an integer.
        """
        end_position = get_end_position(string_end,
                                        text,
                                        position + len(string_start) - 1)
        newline = get_end_position('\n', text, position)
        if newline == -1:
            newline = len(text)
        if end_position == -1:
            _range = SourceRange.from_absolute_position(
                filename,
                AbsolutePosition(file, position))
            raise NoCloseError(string_start, _range)
        if newline > end_position:
            return (SourceRange.from_absolute_position(
                        filename,
                        AbsolutePosition(file, position),
                        AbsolutePosition(file, end_position)),
                    end_position)

    @staticmethod
    def get_singleline_comment(file, filename, text, comment, position):
        """
        Gets Sourcerange of a single-line comment where the start is the
        start of comment and the end is the end of line.

        :param file:
            A tuple of strings, with each string being a line in the file.
        :param filename:
            The name of the file.
        :param comment:
            The string which specifies the comment.
        :position:
            An integer identifying the position where the string started.
        :return:
            A SourceRange object identifying the range of the single-line
            comment and the end_position of the comment as an integer.
        """
        end_position = get_end_position('\n',
                                        text,
                                        position + len(comment) - 1)
        if end_position == -1:
            end_position = len(text) - 1
        return (SourceRange.from_absolute_position(
                    filename,
                    AbsolutePosition(file, position),
                    AbsolutePosition(file, end_position)),
                end_position)


def get_end_position(end_marker, text, position):
    try:
        end_match = next(unescaped_search_for(end_marker, text[position + 1:]))
        end_position = position + end_match.span()[1]
    except StopIteration:
        end_position = -1

    return end_position


class NoCloseError(Exception):

    def __init__(self, annotation, code):
        Exception.__init__(self, annotation + ' has no closure')
        self.code = code



from coalib.bearlib.abstractions.Linter import linter
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable='php',
        output_format='regex',
        output_regex=r'(?P<severity>Parse|Fatal) error: (?P<message>.*) in '
                     r'.* on line (?P<line>\d+)',
        severity_map={'Parse': RESULT_SEVERITY.MAJOR,
                      'Fatal': RESULT_SEVERITY.MAJOR})
class PHPLintBear:
    """
    Checks the code with ``php -l``. This runs it on each file separately.
    """
    LANGUAGES = {'PHP'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ('-l', '-n', '-d', 'display_errors=On', '-d', 'log_errors=Off',
                filename)



"""
This file contains counting conditions for use for count matrix based code
clone detection. (See http://goo.gl/8UuAW5 for general information about the
algorithm.)
"""


from clang.cindex import CursorKind

from coalib.misc.Enum import enum


def is_function_declaration(cursor):
    """
    Checks if the given clang cursor is a function declaration.

    :param cursor: A clang cursor from the AST.
    :return:       A bool.
    """
    return cursor.kind == CursorKind.FUNCTION_DECL


def get_identifier_name(cursor):
    """
    Retrieves the identifier name from the given clang cursor.

    :param cursor: A clang cursor from the AST.
    :return:       The identifier as string.
    """
    return cursor.displayname


def is_literal(cursor):
    """
    :param cursor: A clang cursor from the AST.
    :return:       True if the cursor is a literal of any kind..
    """
    return cursor.kind in [CursorKind.INTEGER_LITERAL,
                           CursorKind.FLOATING_LITERAL,
                           CursorKind.IMAGINARY_LITERAL,
                           CursorKind.STRING_LITERAL,
                           CursorKind.CHARACTER_LITERAL,
                           CursorKind.OBJC_STRING_LITERAL,
                           CursorKind.CXX_BOOL_LITERAL_EXPR,
                           CursorKind.CXX_NULL_PTR_LITERAL_EXPR]


def is_reference(cursor):
    """
    Determines if the cursor is a reference to something, i.e. an identifier
    of a function or variable.

    :param cursor: A clang cursor from the AST.
    :return:       True if the cursor is a reference.
    """
    return cursor.kind in [CursorKind.VAR_DECL,
                           CursorKind.PARM_DECL,
                           CursorKind.DECL_REF_EXPR]


def _stack_contains_kind(stack, kind):
    """
    Checks if a cursor with the given kind is within the stack.

    :param stack: The stack holding a tuple holding the parent cursors and the
                  child number.
    :param kind:  The kind of the cursor to search for.
    :return:      True if the kind was found.
    """
    for elem, _ in stack:
        if elem.kind == kind:
            return True

    return False


def _is_nth_child_of_kind(stack, allowed_nums, kind):
    """
    Checks if the stack contains a cursor which is of the given kind and the
    stack also has a child of this element which number is in the allowed_nums
    list.

    :param stack:        The stack holding a tuple holding the parent cursors
                         and the child number.
    :param allowed_nums: List/iterator of child numbers allowed.
    :param kind:         The kind of the parent element.
    :return:             Number of matches.
    """
    is_kind_child = False
    count = 0
    for elem, child_num in stack:
        if is_kind_child and child_num in allowed_nums:
            count += 1

        if elem.kind == kind:
            is_kind_child = True
        else:
            is_kind_child = False

    return count


def is_function(stack):
    """
    Checks if the cursor on top of the stack is used as a method or as a
    variable.

    :param stack: A stack holding a tuple holding the parent cursors and the
                  child number.
    :return:      True if this is used as a function, false otherwise.
    """
    return _is_nth_child_of_kind(stack, [0], CursorKind.CALL_EXPR) != 0


FOR_POSITION = enum('UNKNOWN', 'INIT', 'COND', 'INC', 'BODY')


def _get_position_in_for_tokens(tokens, position):
    """
    Retrieves the semantic position of the given position in a for loop. It
    operates under the assumptions that the given tokens represent a for loop
    and that the given position is within the tokens.

    :param tokens:   The tokens representing the for loop (clang extent)
    :param position: A tuple holding (line, column) of the position to
                     identify.
    :return:         A FOR_POSITION object indicating where the position is
                     semantically.
    """
    state = FOR_POSITION.INIT
    next_state = state
    opened_brackets = 0
    for token in tokens:
        if token.spelling == ';':
            next_state = state + 1
        elif token.spelling == '(':
            opened_brackets += 1
        elif token.spelling == ')':
            opened_brackets -= 1
            # Closed bracket for for condition, otherwise syntax error by clang
            if opened_brackets == 0:
                next_state = FOR_POSITION.BODY

        if next_state is not state:
            token_position = (token.extent.start.line,
                              token.extent.start.column)
            if position <= token_position:
                return state
            # Last state, if we reach it the position must be in body
            elif next_state == FOR_POSITION.BODY:
                return next_state

            state = next_state

    # We probably have a macro here, clang doesn't preprocess them. I don't see
    # a chance of getting macros parsed right here in the limited time
    # available. For our heuristic approach we'll just not count for loops
    # realized through macros. FIXME: This is not covered in the tests because
    # it contains a known bug that needs to be fixed, that is: macros destroy
    # everything.
    return FOR_POSITION.UNKNOWN  # pragma: no cover


def _get_positions_in_for_loop(stack):
    """
    Investigates all FOR_STMT objects in the stack and checks for each in
    what position the given cursor is.

    :param cursor: The cursor to investigate.
    :param stack:  The stack of parental cursors.
    :return:       A list of semantic FOR_POSITION's within for loops.
    """
    results = []
    for elem, _ in stack:
        if elem.kind == CursorKind.FOR_STMT:
            results.append(_get_position_in_for_tokens(
                elem.get_tokens(),
                (stack[-1][0].location.line, stack[-1][0].location.column)))

    return results


def _get_binop_operator(cursor):
    """
    Returns the operator token of a binary operator cursor.

    :param cursor: A cursor of kind BINARY_OPERATOR.
    :return:       The token object containing the actual operator or None.
    """
    children = list(cursor.get_children())
    operator_min_begin = (children[0].location.line,
                          children[0].location.column)
    operator_max_end = (children[1].location.line,
                        children[1].location.column)

    for token in cursor.get_tokens():
        if (operator_min_begin < (token.extent.start.line,
                                  token.extent.start.column) and
            operator_max_end >= (token.extent.end.line,
                                 token.extent.end.column)):
            return token

    return None  # pragma: no cover


def _stack_contains_operators(stack, operators):
    """
    Checks if one of the given operators is within the stack.

    :param stack:     The stack holding a tuple holding the parent cursors
                      and the child number.
    :param operators: A list of strings. E.g. ["+", "-"]
    :return:          True if the operator was found.
    """
    for elem, _ in stack:
        if elem.kind in [CursorKind.BINARY_OPERATOR,
                         CursorKind.COMPOUND_ASSIGNMENT_OPERATOR]:
            operator = _get_binop_operator(elem)
            # Not known how to reproduce but may be possible when evil macros
            # join the game.
            if operator is None:  # pragma: no cover
                continue

            if operator.spelling in operators:
                return True

    return False


ARITH_BINARY_OPERATORS = ['+', '-', '*', '/', '%', '&', '|']
COMPARISION_OPERATORS = ['==', '<=', '>=', '<', '>', '!=', '&&', '||']
ADV_ASSIGNMENT_OPERATORS = [op + '=' for op in ARITH_BINARY_OPERATORS]
ASSIGNMENT_OPERATORS = ['='] + ADV_ASSIGNMENT_OPERATORS


def in_sum(stack):
    """
    A counting condition returning true if the variable is used in a sum
    statement, i.e. within the operators +, - and their associated compound
    operators.
    """
    return _stack_contains_operators(stack, ['+', '-', '+=', '-='])


def in_product(stack):
    """
    A counting condition returning true if the variable is used in a product
    statement, i.e. within the operators \*, /, % and their associated compound
    operators.
    """
    return _stack_contains_operators(stack, ['*', '/', '%', '*=', '/=', '%='])


def in_binary_operation(stack):
    """
    A counting condition returning true if the variable is used in a binary
    operation, i.e. within the operators \|, & and their associated compound
    operators.
    """
    return _stack_contains_operators(stack, ['&', '|', '&=', '|='])


def member_accessed(stack):
    """
    Returns true if a member of the cursor is accessed or the cursor is the
    accessed member.
    """
    return _stack_contains_kind(stack, CursorKind.MEMBER_REF_EXPR)


# pylint: disabled=unused-argument
def used(stack):
    """
    Returns true.
    """
    return True


def returned(stack):
    """
    Returns true if the cursor on top is used in a return statement.
    """
    return _stack_contains_kind(stack, CursorKind.RETURN_STMT)


def is_inc_or_dec(stack):
    """
    Returns true if the cursor on top is inc- or decremented.
    """
    for elem, _ in stack:
        if elem.kind == CursorKind.UNARY_OPERATOR:
            for token in elem.get_tokens():
                if token.spelling in ['--', '++']:
                    return True

    return False


def is_condition(stack):
    """
    Returns true if the cursor on top is used as a condition.
    """
    return (_is_nth_child_of_kind(stack, [0], CursorKind.WHILE_STMT) != 0 or
            _is_nth_child_of_kind(stack, [0], CursorKind.IF_STMT) != 0 or
            _is_nth_child_of_kind(stack, [0], CursorKind.SWITCH_STMT) != 0 or
            _is_nth_child_of_kind(stack, [0], CursorKind.CASE_STMT) != 0 or
            FOR_POSITION.COND in _get_positions_in_for_loop(stack))


def in_condition(stack):
    """
    Returns true if the cursor on top is in the body of one condition.
    """
    # In every case the first child of IF_STMT is the condition itself
    # (non-NULL) so the second and third child are in the then/else branch
    return (_is_nth_child_of_kind(stack, [1, 2], CursorKind.IF_STMT) == 1 or
            (_is_nth_child_of_kind(stack, [1], CursorKind.SWITCH_STMT) == 1 and
             _is_nth_child_of_kind(stack, [0], CursorKind.CASE_STMT) == 0))


def in_second_level_condition(stack):
    """
    Returns true if the cursor on top is in the body of two nested conditions.
    """
    return _is_nth_child_of_kind(stack, [1, 2], CursorKind.IF_STMT) == 2


def in_third_level_condition(stack):
    """
    Returns true if the cursor on top is in the body of three or more nested
    conditions.
    """
    return _is_nth_child_of_kind(stack, [1, 2], CursorKind.IF_STMT) > 2


def is_assignee(stack):
    """
    Returns true if the cursor on top is assigned something.
    """
    cursor_pos = (stack[-1][0].extent.end.line, stack[-1][0].extent.end.column)
    for elem, _ in stack:
        if (
                elem.kind == CursorKind.BINARY_OPERATOR or
                elem.kind == CursorKind.COMPOUND_ASSIGNMENT_OPERATOR):
            for token in elem.get_tokens():
                token_pos = (token.extent.start.line,
                             token.extent.start.column)
                # This needs to be an assignment and cursor has to be on LHS
                if (
                        token.spelling in ASSIGNMENT_OPERATORS and
                        cursor_pos <= token_pos):
                    return True

    return is_inc_or_dec(stack)


def is_assigner(stack):
    """
    Returns true if the cursor on top is used for an assignment on the RHS.
    """
    cursor_pos = (stack[-1][0].extent.start.line,
                  stack[-1][0].extent.start.column)
    for elem, _ in stack:
        if (
                elem.kind == CursorKind.BINARY_OPERATOR or
                elem.kind == CursorKind.COMPOUND_ASSIGNMENT_OPERATOR):
            for token in elem.get_tokens():
                token_pos = (token.extent.end.line, token.extent.end.column)
                # This needs to be an assignment and cursor has to be on RHS
                # or if we have something like += its irrelevant on which side
                # it is because += reads on both sides
                if (token.spelling in ASSIGNMENT_OPERATORS and (
                        token_pos <= cursor_pos or
                        token.spelling != '=')):
                    return True

    return is_inc_or_dec(stack)


def _loop_level(stack):
    """
    Investigates the stack to determine the loop level.

    :param stack: A stack of clang cursors.
    :return:      An integer representing the level of nested loops.
    """
    positions_in_for = _get_positions_in_for_loop(stack)
    return (positions_in_for.count(FOR_POSITION.INC) +
            positions_in_for.count(FOR_POSITION.BODY) +
            _is_nth_child_of_kind(stack, [1], CursorKind.WHILE_STMT))


def loop_content(stack):
    """
    Returns true if the cursor on top is within a first level loop.
    """
    return _loop_level(stack) == 1


def second_level_loop_content(stack):
    """
    Returns true if the cursor on top is within a second level loop.
    """
    return _loop_level(stack) == 2


def third_level_loop_content(stack):
    """
    Returns true if the cursor on top is within a third (or higher) level loop.
    """
    return _loop_level(stack) > 2


def is_param(stack):
    """
    Returns true if the cursor on top is a parameter declaration.
    """
    return stack[-1][0].kind == CursorKind.PARM_DECL


def is_called(stack):
    """
    Yields true if the cursor is a function that is called. (Function pointers
    are counted too.)
    """
    return (_stack_contains_kind(stack, CursorKind.CALL_EXPR) and
            is_function(stack))


def is_call_param(stack):
    """
    Yields true if the cursor is a parameter to another function.
    """
    return (_stack_contains_kind(stack, CursorKind.CALL_EXPR) and
            not is_function(stack))


condition_dict = {'used': used,
                  'returned': returned,
                  'is_condition': is_condition,
                  'in_condition': in_condition,
                  'in_second_level_condition': in_second_level_condition,
                  'in_third_level_condition': in_third_level_condition,
                  'is_assignee': is_assignee,
                  'is_assigner': is_assigner,
                  'loop_content': loop_content,
                  'second_level_loop_content': second_level_loop_content,
                  'third_level_loop_content': third_level_loop_content,
                  'is_param': is_param,
                  'is_called': is_called,
                  'is_call_param': is_call_param,
                  'in_sum': in_sum,
                  'in_product': in_product,
                  'in_binary_operation': in_binary_operation,
                  'member_accessed': member_accessed}


def counting_condition(value):
    """
    This is a custom converter to convert a setting from coala into counting
    condition function objects for this bear only.

    :param value: An object that can be converted to a list.
    :return:      A list of functions (counting conditions)
    """
    str_list = list(value)
    result_list = []
    for elem in str_list:
        result_list.append(condition_dict.get(elem.lower()))

    return result_list



import argparse
import glob
import os
import shutil
from string import Template
import subprocess
import sys
import time

from bears import VERSION
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.collecting.Importers import iimport_objects
from coalib.parsing.Globbing import glob


def touch(file_name):
    """
    Creates an empty file. An existing file remains untouched.

    :param file_name: Name of the file.
    """
    open(file_name, 'a').close()


def create_file_from_template(template_file, output_file, substitution_dict):
    """
    Creates a file from a template file, using a substitution dict.

    :param template_file:     The template file.
    :param output_file:       The file to be written.
    :param substitution_dict: The dict from which the substitutions are taken.
    """
    with open(template_file) as fl:
        template = fl.read()
    template = Template(template).safe_substitute(**substitution_dict)

    with open(output_file, 'w') as output_handle:
        output_handle.write(template)


def create_file_structure_for_packages(root_folder, file_to_copy, object_name):
    """
    Creates a file structure for the packages to be uploaded. The structure
    will be ``root_folder/object_name/coalaobject_name/object_name.py``.
    Also holds a ``root_folder/object_name/coalaobject_name/__init__.py``
    to make the package importable.

    :param root_folder:  The folder in which the packages are going to be
                         generated.
    :param file_to_copy: The file that is going to be generated the package
                         for.
    :param object_name:  The name of the object that is inside the
                         file_to_copy.
    """
    upload_package_folder = os.path.join(
        root_folder, object_name, 'coala' + object_name)
    os.makedirs(upload_package_folder, exist_ok=True)
    touch(os.path.join(upload_package_folder, '__init__.py'))
    shutil.copyfile(file_to_copy, os.path.join(upload_package_folder,
                                               object_name + '.py'))


def perform_register(path, file_name):
    """
    Register the directory to PyPi, after creating a ``sdist`` and
    a ``bdist_wheel``.

    :param path: The file on which the register should be done.
    """
    subprocess.call(
        [sys.executable, 'setup.py', 'sdist', 'bdist_wheel'], cwd=path)
    subprocess.call(['twine', 'register', '-r', 'pypi', os.path.join(
        path, 'dist', file_name + '.tar.gz')])
    subprocess.call(['twine', 'register', '-r', 'pypi', os.path.join(
        path, 'dist', file_name + '-py3-none-any.whl')])


def perform_upload(path):
    """
    Uploads the directory to PyPi.

    :param path: The folder in which the upload should be done.
    """
    subprocess.call(
        ['twine', 'upload', path + '/dist/*'])


def create_upload_parser():
    """
    Creates a parser for command line arguments.

    :return: Parser arguments.
    """
    parser = argparse.ArgumentParser(
        description='Generates PyPi packages from bears.')
    parser.add_argument('-r', '--register',
                        help='Register the packages on PyPi',
                        action='store_true')
    parser.add_argument('-u', '--upload', help='Upload the packages on PyPi',
                        action='store_true')
    return parser


def main():
    args = create_upload_parser().parse_args()

    os.makedirs(os.path.join('bears', 'upload'), exist_ok=True)

    bear_version = VERSION
    if 'dev' in bear_version:
        bear_version = bear_version[:bear_version.find('dev')] + (
            str(int(time.time())))
    else:
        bear_version = repr(bear_version) + '.' + str(int(time.time()))

    for bear_file_name in sorted(set(glob('bears/**/*Bear.py'))):
        bear_object = next(iimport_objects(
            bear_file_name, attributes='kind', local=True),
            None)
        if bear_object:
            bear_name, _ = os.path.splitext(os.path.basename(bear_file_name))
            create_file_structure_for_packages(
                os.path.join('bears', 'upload'), bear_file_name, bear_name)
            if bear_object.REQUIREMENTS:
                for requirement in bear_object.REQUIREMENTS:
                    if isinstance(requirement, PipRequirement):
                        with open(os.path.join(
                                    'bears', 'upload',
                                    bear_name, 'requirements.txt'),
                                  'a') as reqtxt:
                            reqtxt.write(
                                requirement.package + '=='
                                + requirement.version + '\n')

                if os.path.exists(os.path.join('bears', 'upload',
                                               bear_name, 'requirements.txt')):
                    with open(os.path.join(
                                'bears', 'upload',
                                bear_name, 'MANIFEST.in'), 'w') as manifest:
                        manifest.write('include requirements.txt')

            substitution_dict = {'NAME': repr(bear_name),
                                 'VERSION': bear_version,
                                 'AUTHORS': str(bear_object.AUTHORS),
                                 'AUTHORS_EMAILS':
                                 str(bear_object.AUTHORS_EMAILS),
                                 'MAINTAINERS': str(bear_object.maintainers),
                                 'MAINTAINERS_EMAILS':
                                 str(bear_object.maintainers_emails),
                                 'PLATFORMS': str(bear_object.PLATFORMS),
                                 'LICENSE': str(bear_object.LICENSE),
                                 'LONG_DESCRIPTION': str(bear_object.__doc__),
                                 'BEAR_NAME': bear_name,
                                 'ENTRY': 'coala' + bear_name}

            create_file_from_template(os.path.join('bears', 'setup.py.in'),
                                      os.path.join('bears', 'upload',
                                                   bear_name, 'setup.py'),
                                      substitution_dict)

            bear_dist_name = bear_name + '-' + bear_version
            if args.register:
                perform_register(os.path.join('bears', 'upload', bear_name),
                                 bear_dist_name)
            if args.upload:
                perform_upload(os.path.join('bears', 'upload', bear_name))


if __name__ == '__main__':  # pragma: no cover
    sys.exit(main())



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='goreturns',
        use_stdin=True,
        output_format='corrected',
        result_message='Imports or returns need to be added/removed.')
class GoReturnsBear:
    """
    Proposes corrections of Go code using ``goreturns``.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(
        package='sourcegraph.com/sqs/goreturns', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Security'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ()



import os
import shlex

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.settings.Setting import typed_list


@linter(executable='pylint',
        output_format='regex',
        output_regex=r'L(?P<line>\d+)C(?P<column>\d+): (?P<message>'
                     r'(?P<origin>(?P<severity>[WFECRI])\d+) - .*)',
        severity_map={'F': RESULT_SEVERITY.MAJOR,
                      'E': RESULT_SEVERITY.MAJOR,
                      'W': RESULT_SEVERITY.NORMAL,
                      'C': RESULT_SEVERITY.INFO,
                      'R': RESULT_SEVERITY.INFO,
                      'I': RESULT_SEVERITY.INFO})
class PyLintBear:
    """
    Checks the code with pylint. This will run pylint over each file
    separately.
    """
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('pylint', '1.6')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unused Code', 'Formatting', 'Duplication', 'Security',
                  'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         pylint_disable: typed_list(str)=None,
                         pylint_enable: typed_list(str)=None,
                         pylint_cli_options: str='',
                         pylint_rcfile: str=''):
        """
        :param pylint_disable:     Disable the message, report, category or
                                   checker with the given id(s).
        :param pylint_enable:      Enable the message, report, category or
                                   checker with the given id(s).
        :param pylint_cli_options: Any command line options you wish to be
                                   passed to pylint.
        :param pylint_rcfile:      The rcfile for PyLint.
        """
        args = ('--reports=n', '--persistent=n',
                '--msg-template="L{line}C{column}: {msg_id} - {msg}"')
        if pylint_disable:
            args += ('--disable=' + ','.join(pylint_disable),)
        if pylint_enable:
            args += ('--enable=' + ','.join(pylint_enable),)
        if pylint_cli_options:
            args += tuple(shlex.split(pylint_cli_options))
        if pylint_rcfile:
            args += ('--rcfile=' + pylint_rcfile,)
        else:
            args += ('--rcfile=' + os.devnull,)

        return args + (filename,)



import re

from coalib.bears.LocalBear import LocalBear
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class coalaBear(LocalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/4p1i873ebi9qdfmczn2tvxrm0'
    CAN_DETECT = {'Spelling'}

    def run(self, filename, file):
        """
        Check for the correct spelling of ``coala`` in the file.
        """
        corrected = []
        for line in file:
            wrong_spelling = r'C([oO][aA][lL][aA])'
            corrected += [re.sub(wrong_spelling,
                                 lambda match: 'c' + match.group(1),
                                 line)]
        diffs = Diff.from_string_arrays(file, corrected).split_diff()
        for diff in diffs:
            yield Result(self,
                         '``coala`` is always written with a lower case ``c``',
                         affected_code=(diff.range(filename),),
                         diffs={filename: diff},
                         severity=RESULT_SEVERITY.MAJOR)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='ramllint',
        output_format='regex',
        output_regex=r'(?P<severity>error|warning|info).*\n  (?P<message>.+) '
                     r'\[(?P<origin>.+)\]')
class RAMLLintBear:
    """
    RAML Linter is a static analysis, linter-like, utility that will enforce
    rules on a given RAML document, ensuring consistency and quality.
    Note: Files should not have leading empty lines, else the bear fails to
    identify the problems correctly.
    """

    LANGUAGES = {'RAML'}
    REQUIREMENTS = {NpmRequirement('ramllint', '1.2.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.settings.Setting import typed_list


@linter(executable='cpplint',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+): (?P<message>.+)')
class CPPLintBear:
    """
    Check C++ code for Google's C++ style guide.

    For more information, consult <https://github.com/theandrewdavis/cpplint>.
    """

    LANGUAGES = {'C++'}
    REQUIREMENTS = {PipRequirement('cpplint', '1.3')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         max_line_length: int=79,
                         cpplint_ignore: typed_list(str)=(),
                         cpplint_include: typed_list(str)=()):
        """
        :param max_line_length: Maximum number of characters for a line.
        :param cpplint_ignore:  List of checkers to ignore.
        :param cpplint_include: List of checkers to explicitly enable.
        """
        ignore = ','.join('-'+part.strip() for part in cpplint_ignore)
        include = ','.join('+'+part.strip() for part in cpplint_include)
        return ('--filter=' + ignore + ',' + include,
                '--linelength=' + str(max_line_length),
                filename)



import copy
import math
import os

from munkres import Munkres

from coalib.collecting.Collectors import collect_dirs
from bears.c_languages.codeclone_detection.CountVector import CountVector

# Instantiate globally since this class is holding stateless public methods.
munkres = Munkres()


def exclude_function(count_matrix):
    """
    Determines heuristically whether or not it makes sense for clone
    detection to take this function into account.

    Applied heuristics:
     * Functions with only count vectors with a sum of all unweighted elements
       of lower then 10 are very likely only declarations or empty and to be
       ignored. (Constants are not taken into account.)

    :param count_matrix: A dictionary with count vectors representing all
                         variables for a function.
    :return:             True if the function is useless for evaluation.
    """
    var_list = [cv for cv in count_matrix.values()
                if cv.category is CountVector.Category.reference]
    var_count = len(var_list)
    variable_sum = sum(sum(cv.unweighted) for cv in var_list)
    return (all((cv.category is CountVector.Category.literal or
                 sum(cv.unweighted) < 10)
                for cv in count_matrix.values()) or
            variable_sum < 11 or
            var_count < 2)


def get_count_matrices(count_vector_creator,
                       filenames,
                       progress_callback,
                       base_path,
                       extra_include_paths):
    """
    Retrieves matrices holding count vectors for all variables for all
    functions in the given file.

    :param count_vector_creator: A object with a get_vectors_for_file method
                                 taking a filename as argument.
    :param filenames:            The files to create count vectors for.
    :param progress_callback:    A function with one float argument which is
                                 called after processing each file with the
                                 progress percentage (float) as an argument.
    :param extra_include_paths:  A list containing additional include paths.
    :return:                     A dict holding a tuple of (file, line,
                                 function) as key and as value a dict with
                                 variable names as key and count vector
                                 objects as value.
    """
    result = {}
    maxlen = len(filenames)
    include_paths = collect_dirs([os.path.dirname(base_path) + '/**'])
    include_paths += extra_include_paths

    for i, filename in enumerate(filenames):
        progress_callback(100*(i/maxlen))
        count_dict = count_vector_creator.get_vectors_for_file(filename,
                                                               include_paths)
        for function in count_dict:
            if not exclude_function(count_dict[function]):
                result[(filename,
                        function[0],
                        function[1])] = count_dict[function]

    return result


def pad_count_vectors(cm1, cm2):
    """
    Pads the smaller count matrix with zeroed count vectors.

    :param cm1: First cm. Will not be modified.
    :param cm2: Second cm. Will not be modified.
    :return:    A tuple holding two cms.
    """
    cm1len = len(cm1)
    cm2len = len(cm2)
    if cm1len != cm2len:
        # Copy the smaller matrix as it will be altered
        if cm1len > cm2len:
            cm2 = copy.copy(cm2)
        else:  # make cm1 the larger (or equal) one
            tmp = cm2
            cm2 = copy.copy(cm1)
            cm1 = tmp

        any_count_vector = list(cm1.values())[0]
        # Fill up smaller count matrix with zero vectors. This way no
        # padding is needed later and if count vectors are zero on both
        # side, the difference is zero too which wouldn't be taken into
        # account with simple padding of ones.
        for i in range(len(cm1) - len(cm2)):
            cm2[i] = any_count_vector.create_null_vector(i)

    return cm1, cm2


def relative_difference(difference, maxabs):
    if maxabs == 0:
        return 1
    return difference/maxabs


def average(lst):
    return sum(lst)/len(lst)


def get_difference(matching_iterator,
                   average_calculation,
                   poly_postprocessing,
                   exp_postprocessing):
    """
    Retrieves the difference value for the matched function represented by the
    given matches.

    Postprocessing may be done because small functions are less likely to be
    clones at the same difference value than big functions which may provide a
    better refactoring opportunity for the user.

    :return:                    A difference value between 0 and 1.

    :param matching_iterator:   A list holding tuples of an absolute difference
                                value and a value to normalize the difference
                                into a range of [0, 1].
    :param average_calculation: If set to true this function will take the
                                average of all variable differences as the
                                difference, else it will normalize the
                                function as a whole and thus weighting in
                                variables dependent on their size.
    :param poly_postprocessing: If set to true, the difference value of big
                                function pairs will be reduced using a
                                polynomial approach.
    :param exp_postprocessing:  If set to true, the difference value of big
                                function pairs will be reduced using an
                                exponential approach.
    """
    norm_sum = sum(norm for diff, norm in matching_iterator)

    if average_calculation:
        difference = average([relative_difference(diff, norm)
                              for diff, norm in matching_iterator])
    else:
        difference = relative_difference(
            sum(diff for diff, norm in matching_iterator),
            norm_sum)

    if poly_postprocessing and norm_sum != 0:
        # This function starts at 1 and converges to .75 for norm_sum -> inf
        difference *= (3*norm_sum+1)/(4*norm_sum)
    if exp_postprocessing and norm_sum != 0:
        difference *= math.exp(1-norm_sum)/4 + 0.75

    return difference


def compare_functions(cm1,
                      cm2,
                      average_calculation=False,
                      poly_postprocessing=True,
                      exp_postprocessing=False):
    """
    Compares the functions represented by the given count matrices.

    Postprocessing may be done because small functions are less likely to be
    clones at the same difference value than big functions which may provide a
    better refactoring opportunity for the user.

    :param cm1:                 Count vector dict for the first function.
    :param cm2:                 Count vector dict for the second function.
    :param average_calculation: If set to true the difference calculation
                                function will take the average of all variable
                                differences as the difference, else it will
                                normalize the function as a whole and thus
                                weighting in variables dependent on their size.
    :param poly_postprocessing: If set to true, the difference value of big
                                function pairs will be reduced using a
                                polynomial approach.
    :param exp_postprocessing:  If set to true, the difference value of big
                                function pairs will be reduced using an
                                exponential approach.
    :return:                    The difference between these functions, 0 is
                                identical and 1 is not similar at all.
    """
    assert 0 not in (len(cm1), len(cm2))

    cm1, cm2 = pad_count_vectors(cm1, cm2)

    diff_table = [(cv1,
                   [(cv2, cv1.difference(cv2), cv1.maxabs(cv2))
                    for cv2 in cm2.values()])
                  for cv1 in cm1.values()]

    # The cost matrix holds the difference between the two variables i and
    # j in the i/j field. This is a representation of a bipartite weighted
    # graph with nodes representing the first function on the one side
    # (rows) and the nodes representing the second function on the other
    #  side (columns). The fields in the matrix are the weighted nodes
    # connecting each element from one side to the other.
    cost_matrix = [[relative_difference(difference, maxabs)
                    for cv2, difference, maxabs in lst]
                   for cv1, lst in diff_table]

    # The munkres algorithm will calculate a matching such that the sum of
    # the taken fields is minimal. It thus will associate each variable
    # from one function to one on the other function.
    matching = munkres.compute(cost_matrix)

    return get_difference([(diff_table[x][1][y][1], diff_table[x][1][y][2])
                           for x, y in matching],
                          average_calculation,
                          poly_postprocessing,
                          exp_postprocessing)



from coalib.bearlib.abstractions.Linter import linter


@linter(executable='shellcheck', output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): '
                     r'(?P<severity>error|warning|info): (?P<message>.+)')
class ShellCheckBear:
    """
    Check bash/shell scripts for syntactical problems (with understandable
    messages), semantical problems as well as subtle caveats and pitfalls.

    A gallery of bad code that can be detected is available at
    <https://github.com/koalaman/shellcheck/blob/master/README.md>.
    """

    LANGUAGES = {'sh', 'bash', 'ksh', 'dash'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Security', 'Undefined Element', 'Unused Code'}

    @staticmethod
    def create_arguments(filename, file, config_file, shell: str='sh'):
        """
        :param shell: Target shell being used.
        """
        return '--f', 'gcc', '-s', shell, filename



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='go',
        use_stdout=False,
        use_stderr=True,
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+): (?P<message>.*)')
class GoVetBear:
    """
    Analyze Go code and raise suspicious constructs, such as printf calls
    whose arguments do not correctly match the format string, useless
    assignments, common mistakes about boolean operations, unreachable code,
    etc.

    This is done using the ``vet`` command. For more information visit
    <https://golang.org/cmd/vet/>.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(package='golang.org/cmd/vet', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Unused Code', 'Smell', 'Unreachable Code'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return 'vet', filename



from coalib.bearlib.abstractions.Linter import linter

from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='pycodestyle',
        output_format='regex',
        output_regex=r'(?P<line>\d+) (?P<column>\d+) '
                     r'(?P<message>(?P<origin>\S+).*)')
class PycodestyleBear:
    """
    A wrapper for the tool ``pycodestyle`` formerly known as ``pep8``.
    """
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('pycodestyle', '2.2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @staticmethod
    def create_arguments(
            filename, file, config_file,
            pycodestyle_ignore: str='',
            pycodestyle_select: str='',
            max_line_length: int=79):
        """
        :param pycodestyle_ignore:
            Comma separated list of errors to ignore.
            See ``pydocstyle`` documentation for a complete list of errors.
        :param pycodestyle_select:
            Comma separated list of errors to detect. If given only
            these errors are going to be detected.
            See ``pydocstyle`` documentation for a complete list of errors.
        :param max_line_length:
            Limit lines to this length.
        """
        arguments = [r"--format='%(row)d %(col)d %(code)s %(text)s'"]

        if pycodestyle_ignore:
            arguments.append('--ignore=' + pycodestyle_ignore)

        if pycodestyle_select:
            arguments.append('--select=' + pycodestyle_select)

        arguments.append('--max-line-length=' + str(max_line_length))

        arguments.append(filename)

        return arguments



import shlex

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GoRequirement import GoRequirement


@linter(executable='golint',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+): (?P<message>.*)')
class GoLintBear:
    """
    Checks the code using ``golint``. This will run golint over each file
    separately.
    """
    LANGUAGES = {'Go'}
    REQUIREMENTS = {GoRequirement(
        package='github.com/golang/lint/golint', flag='-u')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         golint_cli_options: str=''):
        """
        :param golint_cli_options: Any other flags you wish to pass to golint.
        """
        args = ()
        if golint_cli_options:
            args += tuple(shlex.split(golint_cli_options))
        return args + (filename,)



import re

from coalib.bearlib import deprecate_settings
from coalib.bears.LocalBear import LocalBear
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


class MatlabIndentationBear(LocalBear):
    LANGUAGES = {'Matlab', 'Octave'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @deprecate_settings(indent_size='tab_width')
    def run(self, filename, file, indent_size: int=2):
        """
        This bear features a simple algorithm to calculate the right
        indentation for Matlab/Octave code. However, it will not handle hanging
        indentation or conditions ranging over several lines yet.

        :param indent_size: Number of spaces per indentation level.
        """
        new_file = tuple(self.reindent(file, indent_size))

        if new_file != tuple(file):
            wholediff = Diff.from_string_arrays(file, new_file)
            for diff in wholediff.split_diff():
                yield Result(
                    self,
                    'The indentation could be changed to improve readability.',
                    severity=RESULT_SEVERITY.INFO,
                    affected_code=(diff.range(filename),),
                    diffs={filename: diff})

    @staticmethod
    def reindent(file, indentation):
        indent, nextindent = 0, 0
        for line_nr, line in enumerate(file, start=1):
            indent = nextindent
            indent, nextindent = MatlabIndentationBear.get_indent(line,
                                                                  indent,
                                                                  nextindent)
            stripped = line.lstrip()
            if stripped:
                yield indent*indentation*' ' + stripped
            else:
                yield line

    @staticmethod
    def get_indent(line, indent, nextindent):
        ctrlstart = r'\s*(function|if|while|for|switch)'
        ctrlcont = r'\s*(elseif|else|case|catch|otherwise)'
        ctrlend = r'\s*(end|endfunction|endif|endwhile|endfor|endswitch)'
        if re.match(ctrlstart, line) is not None:
            return indent, nextindent+1
        elif re.match(ctrlcont, line) is not None:
            return indent-1, nextindent
        elif re.match(ctrlend, line) is not None:
            return indent-1, nextindent-1
        else:
            return indent, indent



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='verilator',
        output_format='regex',
        use_stderr=True,
        output_regex=r'\%(?:(?P<severity>Error|Warning.*?).*?):'
                     r'.+?:(?P<line>.+?): (?P<message>.+)')
class VerilogLintBear:
    """
    Analyze Verilog code using ``verilator`` and checks for all lint
    related and code style related warning messages. It supports the
    synthesis subset of Verilog, plus initial statements, proper
    blocking/non-blocking assignments, functions, tasks.

    It also warns about unused code when a specified signal is never sinked,
    and unoptimized code due to some construct, with which the
    optimization of the specified signal or block is disabled.

    This is done using the ``--lint-only`` command. For more information visit
    <http://www.veripool.org/projects/verilator/wiki/Manual-verilator>.
    """
    LANGUAGES = {'Verilog'}
    REQUIREMENTS = {DistributionRequirement(apt_get='verilator')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/45275'
    CAN_DETECT = {'Formatting', 'Code Simplification', 'Syntax', 'Unused Code'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--lint-only', filename



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement


@linter(executable='puppet-lint',
        output_format='regex',
        output_regex=r'(?P<line>\d+):(?P<column>\d+):'
                     r'(?P<severity>warning|error):(?P<message>.+)')
class PuppetLintBear:
    '''
    Check and correct puppet configuration files using ``puppet-lint``.

    See <http://puppet-lint.com/> for details about the tool.
    '''

    LANGUAGES = {'Puppet'}
    REQUIREMENTS = {GemRequirement('puppet-lint', '2')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/87751'
    CAN_FIX = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return ('--log-format', '%{line}:%{column}:%{kind}:%{message}',
                filename)



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)


@linter(executable='chktex',
        output_format='regex',
        output_regex=r'(?P<severity>Error|Warning) \d+ in .+ line '
                     r'(?P<line>\d+): (?P<message>.*)')
class LatexLintBear:
    """
    Checks the code with ``chktex``.
    """
    LANGUAGES = {'Tex'}
    REQUIREMENTS = {DistributionRequirement(apt_get='chktex')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return (
            '--format',
            '%k %n in {0} line %l: %m!n'.format(filename),
            filename,
        )



from shutil import which

from coalib.bearlib.abstractions.Linter import linter


@linter(executable='perl',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<severity>.*) - (?P<message>.*)')
class VHDLLintBear:
    """
    Check VHDL code for common codestyle problems.

    Rules include:

     * Signals, variables, ports, types, subtypes, etc. must be lowercase.
     * Constants and generics must be uppercase.
     * Entities, architectures and packages must be "mixedcase" (may be 100%
       uppercase, but not 100% lowercase).
     * Ports must be suffixed using _i, _o or _io denoting its kind.
     * Labels must be placed in a separated line. Exception: component
       instantiation.
     * End statements must be documented indicating what are finishing.
     * Buffer ports are forbidden.
     * VHDL constructions of the "entity xxxx is" and similars must be in one
       line. You can't put "entity xxxxx" in one line and "is" in another.
     * No more than one VHDL construction is allowed in one line of code.

    See <http://fpgalibre.sourceforge.net/ingles.html#tp46> for more
    information.
    """

    LANGUAGES = {'VHDL'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    @classmethod
    def check_prerequisites(cls):  # pragma: no cover
        if which('perl') is None:
            return 'perl is not installed.'
        elif which('bakalint.pl') is None:
            return ('bakalint is missing. Download it from '
                    '<http://fpgalibre.sourceforge.net/ingles.html#tp46> and '
                    'put it into your PATH.')
        else:
            return True

    @staticmethod
    def create_arguments(filename, file, config_file):
        return which('bakalint.pl'), '--input', filename



from coalib.bearlib import deprecate_settings
from coalib.bearlib.spacing.SpacingHelper import SpacingHelper
from coalib.bears.LocalBear import LocalBear
from coalib.results.Diff import Diff
from coalib.results.Result import Result


class SpaceConsistencyBear(LocalBear):
    LANGUAGES = {'All'}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_FIX = {'Formatting'}

    @deprecate_settings(indent_size='tab_width')
    def run(self,
            filename,
            file,
            use_spaces: bool,
            allow_trailing_whitespace: bool=False,
            indent_size: int=SpacingHelper.DEFAULT_TAB_WIDTH,
            enforce_newline_at_EOF: bool=True):
        '''
        Check and correct spacing for all textual data. This includes usage of
        tabs vs. spaces, trailing whitespace and (missing) newlines before
        the end of the file.

        :param use_spaces:                True if spaces are to be used instead
                                          of tabs.
        :param allow_trailing_whitespace: Whether to allow trailing whitespace
                                          or not.
        :param indent_size:               Number of spaces per indentation
                                          level.
        :param enforce_newline_at_EOF:    Whether to enforce a newline at the
                                          End Of File.
        '''
        spacing_helper = SpacingHelper(indent_size)
        result_texts = []
        additional_info_texts = []

        for line_number, line in enumerate(file, start=1):
            replacement = line

            if enforce_newline_at_EOF:
                # Since every line contains at the end at least one \n, only
                # the last line could potentially not have one. So we don't
                # need to check whether the current line_number is the last
                # one.
                if replacement[-1] != '\n':
                    replacement += '\n'
                    result_texts.append('No newline at EOF.')
                    additional_info_texts.append(
                        "A trailing newline character ('\\n') is missing from "
                        'your file. '
                        '<http://stackoverflow.com/a/5813359/3212182> gives '
                        'more information about why you might need one.')

            if not allow_trailing_whitespace:
                replacement = replacement.rstrip(' \t\n') + '\n'
                if replacement != line.rstrip('\n') + '\n':
                    result_texts.append('Trailing whitespaces.')
                    additional_info_texts.append(
                        'Your source code contains trailing whitespaces. '
                        'Those usually have no meaning. Please consider '
                        'removing them.')

            if use_spaces:
                pre_replacement = replacement
                replacement = spacing_helper.replace_tabs_with_spaces(
                    replacement)
                if replacement != pre_replacement:
                    result_texts.append('Tabs used instead of spaces.')
            else:
                pre_replacement = replacement
                replacement = spacing_helper.replace_spaces_with_tabs(
                    replacement)
                if replacement != pre_replacement:
                    result_texts.append('Spaces used instead of tabs.')

            if len(result_texts) > 0:
                diff = Diff(file)
                diff.change_line(line_number, line, replacement)
                inconsistencies = ''.join('\n- ' + string
                                          for string in result_texts)
                yield Result.from_values(
                    self,
                    'Line contains following spacing inconsistencies:'
                    + inconsistencies,
                    diffs={filename: diff},
                    file=filename,
                    line=line_number,
                    additional_info='\n\n'.join(additional_info_texts))
                result_texts = []
                additional_info_texts = []



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.PipRequirement import PipRequirement


@linter(executable='cppclean',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<message>.*)')
class CPPCleanBear:
    """
    Find problems in C++ source code that slow down development in large code
    bases. This includes finding unused code, among other features.

    Read more about available routines at
    <https://github.com/myint/cppclean#features>.
    """

    LANGUAGES = {'C++'}
    REQUIREMENTS = {PipRequirement('cppclean', '0.12.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Smell', 'Unused Code', 'Security'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return filename,



import platform

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.DistributionRequirement import (
    DistributionRequirement)
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY


@linter(executable=('perlcritic.bat'
                    if platform.system() == 'Windows' else
                    'perlcritic'),
        output_format='regex',
        output_regex=r'(?P<message>.+) at '
                     r'line (?P<line>\d+), '
                     r'column (?P<column>\d+)\. '
                     r'(?P<origin>.+) '
                     r'\(Severity: (?P<severity>\d+)\)',
        severity_map={'1': RESULT_SEVERITY.MAJOR,
                      '2': RESULT_SEVERITY.MAJOR,
                      '3': RESULT_SEVERITY.NORMAL,
                      '4': RESULT_SEVERITY.NORMAL,
                      '5': RESULT_SEVERITY.INFO})
class PerlCriticBear:
    """
    Check the code with perlcritic. This will run perlcritic over
    each of the files seperately.
    """

    LANGUAGES = {'Perl'}
    REQUIREMENTS = {DistributionRequirement(apt_get='perl'),
                    DistributionRequirement(apt_get='libperl-critic-perl')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting', 'Code Simplification'}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         perlcritic_profile: str=''):
        """
        :param perlcritic_profile: Location of the perlcriticrc config file.
        """
        args = ('--no-color',)
        if perlcritic_profile:
            args += ('--profile', perlcritic_profile)
        return args + (filename,)



from coalib.bears.LocalBear import LocalBear
from bears.general.AnnotationBear import AnnotationBear
from coalib.results.Diff import Diff
from coalib.results.HiddenResult import HiddenResult
from coalib.results.Result import Result


class QuotesBear(LocalBear):

    BEAR_DEPS = {AnnotationBear}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Formatting'}

    def correct_single_line_str(self, filename, file, sourcerange,
                                preferred_quotation):
        """
        Corrects a given single line string assuming it does not use the
        preferred quotation. If the preferred quotation mark is used inside the
        string, no correction will be made.

        This function will yield one or no Result objects.

        :param filename:
            The filename of the file to correct the line in.
        :param file:
            The file contents as list of lines.
        :param sourcerange:
            The sourcerange indicating where to find the string.
        :param preferred_quotation:
            ``'`` or ``"`` respectively.
        """
        str_contents = file[sourcerange.start.line - 1][
                       sourcerange.start.column:sourcerange.end.column-1]

        if preferred_quotation in str_contents:
            return

        before = file[sourcerange.start.line - 1][:sourcerange.start.column-1]
        after = file[sourcerange.end.line - 1][sourcerange.end.column:]

        replacement = (before + preferred_quotation + str_contents +
                       preferred_quotation + after)

        diff = Diff(file)
        diff.change_line(sourcerange.start.line,
                         file[sourcerange.start.line - 1],
                         replacement)
        yield Result(self, 'You do not use the preferred quotation marks.',
                     diff.affected_code(filename), diffs={filename: diff})

    def run(self, filename, file, dependency_results,
            preferred_quotation: str='"'):
        """
        Checks and corrects your quotation style.

        For all single line strings, this bear will correct the quotation to
        your preferred quotation style if that kind of quote is not included
        within the string. Multi line strings are not supported.

        :param preferred_quotation: Your preferred quotation character, e.g.
                                    ``"`` or ``'``.
        """
        if not isinstance(dependency_results[AnnotationBear.name][0],
                          HiddenResult):
            return
        if isinstance(dependency_results[AnnotationBear.name][0].contents,
                      str):
            self.err(dependency_results[AnnotationBear.name][0].contents)
            return

        ranges = dependency_results[AnnotationBear.name][0].contents['strings']

        for string_range in ranges:
            if (file[string_range.start.line-1][string_range.start.column-1] ==
                    preferred_quotation):
                continue

            if string_range.start.line == string_range.end.line:
                yield from self.correct_single_line_str(
                    filename, file, string_range, preferred_quotation)



import radon.complexity
import radon.visitors

from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Result import Result
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.SourceRange import SourceRange
from coalib.settings.Setting import typed_list


class RadonBear(LocalBear):
    LANGUAGES = {'Python', 'Python 2', 'Python 3'}
    REQUIREMENTS = {PipRequirement('radon', '1.4.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Complexity'}

    def run(self, filename, file,
            radon_ranks_info: typed_list(str)=(),
            radon_ranks_normal: typed_list(str)=('C', 'D'),
            radon_ranks_major: typed_list(str)=('E', 'F')):
        """
        Uses radon to compute complexity of a given file.

        :param radon_ranks_info:   The ranks (given by radon) to
                                   treat as severity INFO.
        :param radon_ranks_normal: The ranks (given by radon) to
                                   treat as severity NORMAL.
        :param radon_ranks_major:  The ranks (given by radon) to
                                   treat as severity MAJOR.
        """
        severity_map = {
            RESULT_SEVERITY.INFO: radon_ranks_info,
            RESULT_SEVERITY.NORMAL: radon_ranks_normal,
            RESULT_SEVERITY.MAJOR: radon_ranks_major
        }
        for visitor in radon.complexity.cc_visit(''.join(file)):
            rank = radon.complexity.cc_rank(visitor.complexity)
            severity = None
            for result_severity, rank_list in severity_map.items():
                if rank in rank_list:
                    severity = result_severity
            if severity is None:
                continue

            col = visitor.col_offset if visitor.col_offset else None
            visitor_range = SourceRange.from_values(
                filename, visitor.lineno, col, visitor.endline)
            message = '{} has a cyclomatic complexity of {}'.format(
                visitor.name, rank)

            yield Result(self, message, severity=severity,
                         affected_code=(visitor_range,))



import shutil

from guess_language import guess_language

from coalib.bearlib import deprecate_settings
from coalib.bears.LocalBear import LocalBear
from dependency_management.requirements.PipRequirement import PipRequirement
from coalib.results.Diff import Diff
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange
from coalib.settings.Setting import typed_list


class LanguageToolBear(LocalBear):
    LANGUAGES = {'Natural Language'}
    REQUIREMENTS = {PipRequirement('guess-language-spirit', '0.5.2'),
                    PipRequirement('language-check', '0.8')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Spelling', 'Grammar'}

    @classmethod
    def check_prerequisites(cls):
        if shutil.which('java') is None:
            return 'java is not installed.'
        else:
            try:
                from language_check import LanguageTool, correct
                LanguageTool
                correct
                return True
            except ImportError:  # pragma: no cover
                return 'Please install the `language-check` pip package.'

    @deprecate_settings(language='locale')
    def run(self,
            filename,
            file,
            language: str='auto',
            languagetool_disable_rules: typed_list(str)=()):
        '''
        Checks the code with LanguageTool.

        :param language:                   A locale representing the language
                                           you want to have checked. If set to
                                           'auto' the language is guessed.
                                           If the language cannot be guessed,
                                           'en-US' is used.
        :param languagetool_disable_rules: List of rules to disable checks for.
        '''
        # Defer import so the check_prerequisites can be run without
        # language_check being there.
        from language_check import LanguageTool, correct

        joined_text = ''.join(file)
        language = (guess_language(joined_text)
                    if language == 'auto' else language)
        language = 'en-US' if not language else language

        tool = LanguageTool(language, motherTongue='en_US')
        tool.disabled.update(languagetool_disable_rules)

        matches = tool.check(joined_text)
        for match in matches:
            if not match.replacements:
                diffs = None
            else:
                replaced = correct(joined_text, [match]).splitlines(True)
                diffs = {filename:
                         Diff.from_string_arrays(file, replaced)}

            rule_id = match.ruleId
            if match.subId is not None:
                rule_id += '[{}]'.format(match.subId)

            message = match.msg + ' (' + rule_id + ')'
            source_range = SourceRange.from_values(filename,
                                                   match.fromy+1,
                                                   match.fromx+1,
                                                   match.toy+1,
                                                   match.tox+1)
            yield Result(self, message, diffs=diffs,
                         affected_code=(source_range,))



import json

from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.Diff import Diff
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result


@linter(executable='eslint',
        use_stdin=True,
        use_stderr=True)
class ESLintBear:
    """
    Check JavaScript and JSX code for style issues and semantic errors.

    Find out more at <http://eslint.org/docs/rules/>.
    """

    LANGUAGES = {'JavaScript', 'JSX'}
    REQUIREMENTS = {NpmRequirement('eslint', '2'),
                    NpmRequirement('babel-eslint', '6'),
                    NpmRequirement('eslint-plugin-import', '1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/38739'
    CAN_DETECT = {'Syntax'}
    CAN_FIX = {'Formatting'}

    severity_map = {2: RESULT_SEVERITY.MAJOR,
                    1: RESULT_SEVERITY.NORMAL,
                    0: RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file,
                         eslint_config: str=''):
        """
        :param eslint_config: The location of the .eslintrc config file.
        """
        args = (
            '--no-ignore',
            '--no-color',
            '-f=json',
            '--stdin',
            '--stdin-filename=' + filename,
        )

        if eslint_config:
            args += ('--config', eslint_config)
        else:
            args += ('--config', config_file)

        return args

    @staticmethod
    def generate_config(filename, file):
        return '{"extends": "eslint:recommended"}'

    def process_output(self, output, filename, file):
        if output[1]:
            self.warn('While running {0}, some issues were found:'
                      .format(self.__class__.__name__))
            self.warn(output[1])

        if not file or not output[0]:
            return

        output = json.loads(output[0])
        lines = ''.join(file)

        assert len(output) == 1

        for result in output[0]['messages']:
            if 'fix' not in result:
                diffs = None
            else:
                fix = result['fix']
                start, end = fix['range']
                replacement_text = fix['text']
                new_output = lines[:start] + replacement_text + lines[end:]
                diffs = {filename: Diff.from_string_arrays(
                    lines.splitlines(True), new_output.splitlines(True))}

            origin = (
                '{class_name} ({rule})'.format(class_name=type(self).__name__,
                                               rule=result['ruleId'])
                if result['ruleId'] is not None else self)
            yield Result.from_values(
                origin=origin, message=result['message'],
                file=filename, line=result['line'], diffs=diffs,
                severity=self.severity_map[result['severity']])



import json

from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement
from coalib.results.RESULT_SEVERITY import RESULT_SEVERITY
from coalib.results.Result import Result
from coala_utils.param_conversion import negate


@linter(executable='coffeelint',
        use_stdin=True)
class CoffeeLintBear:
    """
    Check CoffeeScript code for a clean and consistent style.

    For more information about coffeelint, visit <http://www.coffeelint.org/>.
    """

    LANGUAGES = {'CoffeeScript'}
    REQUIREMENTS = {NpmRequirement('coffeelint', '1')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Syntax', 'Formatting', 'Smell', 'Complexity', 'Duplication'}

    severity_map = {'warn': RESULT_SEVERITY.NORMAL,
                    'error': RESULT_SEVERITY.MAJOR,
                    'ignore': RESULT_SEVERITY.INFO}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--reporter=raw', '--stdin', '-f', config_file

    @staticmethod
    @deprecate_settings(indent_size='tab_width',
                        allow_increment=(
                            'no_decr_or_incrementation_operators', negate),
                        allow_no_parameters=(
                            'no_empty_parameter_list', negate),
                        allow_empty_functions=('no_empty_functions', negate),
                        allow_this_statements=('no_this', negate),
                        allow_implicit_parentheses=(
                            'no_implicit_parentheses', negate),
                        allow_interpolation_in_single_quotes=(
                            'no_interpolation_in_single_quotes', negate),
                        allow_stand_alone_at_sign=(
                            'no_stand_alone_at_sign', negate),
                        allow_throwing_strings=(
                            'disable_throwing_strings', negate),
                        allow_unnecessary_double_quotes=(
                            'no_unnecessary_double_quotes', negate),
                        allow_bitwise_operators=(
                            'use_english_operator', negate),
                        force_braces='no_implicit_braces')
    def generate_config(filename, file,
                        max_line_length: int=79,
                        max_line_length_affect_comments: bool=True,
                        space_before_and_after_arrow: bool=True,
                        check_braces_spacing: bool=False,
                        braces_spacing_width: int=1,
                        spacing_in_empty_braces: int=0,
                        class_naming_camelCase: bool=True,
                        spaces_before_and_after_colon: bool=False,
                        spaces_before_colon: int=0,
                        spaces_after_colon: int=1,
                        enforce_newline_at_EOF: bool=True,
                        use_spaces: bool=True,
                        indent_size: int=2,
                        number_of_newlines_after_classes: int=2,
                        prohibit_embedding_javascript_snippet: bool=True,
                        force_braces: bool=False,
                        allow_implicit_parentheses: bool=True,
                        allow_interpolation_in_single_quotes: bool=True,
                        allow_stand_alone_at_sign: bool=False,
                        allow_throwing_strings: bool=False,
                        allow_trailing_semicolons: bool=False,
                        allow_trailing_whitespaces: bool=False,
                        allow_unnecessary_double_quotes: bool=True,
                        allow_bitwise_operators: bool=True,
                        spaces_around_operators: bool=True,
                        space_after_comma: bool=True,
                        cyclomatic_complexity: int=0,
                        prevent_duplicate_keys: bool=True,
                        consistent_line_endings_style: str='',
                        allow_this_statements: bool=True,
                        allow_increment: bool=True,
                        allow_no_parameters: bool=True,
                        allow_empty_functions: bool=False,
                        enforce_parentheses_on_non_empty_constructors:
                            bool=True
                        ):
        """
        :param max_line_length:
            Maximum number of characters per line.
        :param max_line_length_affect_comments:
            Determines if ``max_line_length`` should also affects comments or
            not.
        :param space_before_and_after_arrow:
            Determines if spaces should be used before and after the arrow.
        :param check_braces_spacing:
            Checks if proper spacing is used inside curly braces.
        :param braces_spacing_width:
            Determines the number of blank spaces after the opening ``{`` and
            before the closing brace ``}`` given that there is something within
            the braces.
        :param spacing_in_empty_braces:
            Determines the number of blank spaces after the opening ``{`` and
            before the closing brace ``}`` given empty content.
        :param class_naming_camelCase:
            Checks whether the classes name should be in camel-case or not.
        :param spaces_before_and_after_colon:
            Checks the number of spaces before and after colon.
        :param spaces_before_colon:
            Determines the number of blank spaces before colon when
            ``spaces_before_and_after_colon == True``.
        :param spaces_after_colon:
            Determines the number of space after colon when
            ``spaces_before_and_after_colon == True``.
        :param enforce_newline_at_EOF:
            Checks if the file ends with a single newline.
        :param use_spaces:
            Forbids tabs in indentation and applies two spaces for this
            purpose.
        :param indent_size:
            Number of spaces per indentation level.
        :param number_of_newlines_after_classes:
            Determines the number of newlines that separate the class
            definition and the rest of the code.
        :param prohibit_embedding_javascript_snippet:
            Prevents some JavaScript elements like ``eval`` to affect
            CoffeeScript.
        :param force_braces:
            Prohibits implicit braces when declaring object literals.

            Example: If ``force_braces = True`` then
            ```
            1:2, 3:4
            ```
            is prohibited, whereas
            ```
            {1:2, 3:4}
            ```
            is accepted.
        :param allow_implicit_parentheses:
            Allows implicit parentheses.
        :param allow_interpolation_in_single_quotes:
            Allows string interpolation in a single quoted string.

            Example: If ``allow_interpolation_in_single_quotes = False`` then
            ```
            f = '#{bar}'
            ```
            is prohibited, whereas
            ```
            f = "#{bar}"
            ```
            is correct.
        :param allow_stand_alone_at_sign:
            Allows the use of stand alone  ``@``.

            Example: If ``allow_stand_alone_at_sign = False``
            ```
            @ notok
            not(@).ok
            @::
            ```
            are prohibited, whereas
            ```
            @alright
            @(fn)
            @ok()
            @[ok]
            @ok()
            ```
            are accepted.
        :param allow_throwing_strings:
            Allows throwing string literals or interpolation.

            Example: If ``allow_throwing_strings = False``
            ```
            throw 'my error'
            throw "#{1234}"
            ```
            will not be permitted.
        :param allow_trailing_semicolons:
            Prohibits trailing semicolons when ``False`` since they are
            not useful. The semicolon is meaningful only if there's another
            instruction on the same line.

            Example: If ``allow_trailing_semicolon = False``
            ```
            x = '1234'; console.log(x)
            ```
            Here the semicolon is meaningful.
            ```
            alert('end of line');
            ```
            This semicolon is redundant.
        :param allow_trailing_whitespaces:
            Checks whether to allow trailing whitespacess in the code or not.
        :param allow_unnecessary_double_quotes:
            Allows enclosing strings in double quotes.
        :param allow_bitwise_operators:
            Determines if ``and``, ``or``, ``is`` and ``isnt`` should be used
            instead of ``&&``, ``||``, ``==`` and ``!=``.
        :param spaces_around_operators:
            Enforces that operators have spaces around them.
        :param space_after_comma:
            Checks if there is a blank space after commas.
        :param cyclomatic_complexity:
            Maximum cyclomatic complexity of the file.
        :param prevent_duplicate_keys:
            Prevents defining duplicate keys in object literals and classes.
        :param enforce_parentheses_on_non_empty_constructors:
            Requires constructors with parameters to include parentheses.

            Example:
            ```
            class Foo
            # Warn about missing parentheses here
            a = new Foo
            b = new bar.foo.Foo
            # The parentheses make it clear no parameters are intended
            c = new Foo()
            d = new bar.foo.Foo()
            e = new Foo 1, 2
            f = new bar.foo.Foo 1, 2
            ```
        :param consistent_line_endings_style:
            The option to ``line_endings``, its value is either ``unix`` or
            ``windows``.
        :param allow_this_statements:
            Allows the use of ``this``. ``@`` should be used if ``False``.
        :param allow_increment:
            Allows the use of increment and decrement arithmetic operators.
        :param allow_no_parameters:
            Allows empty parameter lists in function definitions.
        :param allow_empty_functions:
            Allows declaring empty functions.
        """
        coffee_configs = {'max_line_length':
                          {'value': max_line_length,
                           'level': 'error',
                           'limitComments':
                               max_line_length_affect_comments}}
        coffee_configs['arrow_spacing'] = (
            {'level': 'error' if space_before_and_after_arrow else 'ignore'})
        if check_braces_spacing:
            coffee_configs['braces_spacing'] = (
                {'level': 'error',
                 'spaces': braces_spacing_width,
                 'empty_object_spaces': spacing_in_empty_braces})
        if class_naming_camelCase:
            coffee_configs['camel_case_classes'] = {'level': 'error'}
        if spaces_before_and_after_colon:
            coffee_configs['colon_assignment_spacing'] = (
                {'level': 'error',
                 'spacing': {'left': spaces_before_colon,
                             'right': spaces_after_colon}})
        coffee_configs['eol_last'] = (
            {'level': 'error' if enforce_newline_at_EOF else 'ignore'})
        coffee_configs['newlines_after_classes'] = (
            {'value': number_of_newlines_after_classes,
             'level': 'error'})
        coffee_configs['no_backticks'] = (
            {'level': 'error'
                if prohibit_embedding_javascript_snippet else 'ignore'})
        if force_braces:
            coffee_configs['no_implicit_braces'] = (
                {'level': 'error', 'strict': True})
        if not allow_implicit_parentheses:
            coffee_configs['no_implicit_parens'] = (
                {'strict': True, 'level': 'error'})
        coffee_configs['no_interpolation_in_single_quotes'] = (
            {'level': 'error'
                if not allow_interpolation_in_single_quotes else 'ignore'})
        if not allow_stand_alone_at_sign:
            coffee_configs['no_stand_alone_at'] = {'level': 'error'}
        if use_spaces:
            coffee_configs['no_tabs'] = {'level': 'error'}
        coffee_configs['indentation'] = (
            {'value': indent_size, 'level': 'error'})
        coffee_configs['no_throwing_strings'] = (
            {'level': 'error' if not allow_throwing_strings else 'ignore'})
        coffee_configs['no_trailing_semicolons'] = (
            {'level': 'error' if not allow_trailing_semicolons else 'ignore'})
        if not allow_trailing_whitespaces:
            coffee_configs['no_trailing_whitespace'] = (
                {'level': 'error',
                 'allowed_in_comments': True,
                 'allowed_in_empty_lines': True})
        if not allow_unnecessary_double_quotes:
            coffee_configs['no_unnecessary_double_quotes'] = {'level': 'error'}
        if not allow_bitwise_operators:
            coffee_configs['prefer_english_operator'] = (
                {'level': 'error', 'doubleNotLevel': 'ignore'})
        if spaces_around_operators:
            coffee_configs['space_operators'] = {'level': 'error'}
        if space_after_comma:
            coffee_configs['spacing_after_comma'] = {'level': 'warn'}
        coffee_configs['cyclomatic_complexity'] = (
                {'value': cyclomatic_complexity,
                 'level': ('error' if cyclomatic_complexity else 'ignore')})
        coffee_configs['duplicate_key'] = (
            {'level': 'error' if prevent_duplicate_keys else 'ignore'})
        if enforce_parentheses_on_non_empty_constructors:
            coffee_configs['non_empty_constructor_needs_parens'] = (
                {'level': 'error'})
        if consistent_line_endings_style:
            coffee_configs['line_endings'] = (
                {'level': 'error', 'value': consistent_line_endings_style})
        if not allow_this_statements:
            coffee_configs['no_this'] = {'level': 'error'}
        if not allow_increment:
            coffee_configs['no_plusplus'] = {'level': 'error'}
        coffee_configs['no_empty_param_list'] = (
            {'level': 'error' if not allow_no_parameters else 'ignore'})
        coffee_configs['no_empty_functions'] = (
            {'level': 'error' if not allow_empty_functions else 'ignore'})

        return json.dumps(coffee_configs)

    def process_output(self, output, filename, file):
        output = json.loads(output)

        assert len(output) == 1, (
            'More than 1 file parsed, something went wrong')
        for item in tuple(output.values())[0]:
            yield Result.from_values(
                origin='{} ({})'.format(self.name, item['rule']),
                message=item['message'],
                file=filename,
                line=item.get('lineNumber', None),
                end_line=item.get('lineNumberEnd', None),
                severity=self.severity_map[item['level']],
                additional_info=item.get('description',
                                         item.get('context', '')))



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.CabalRequirement import (
    CabalRequirement)


@linter(executable='ghc-mod',
        output_format='regex',
        output_regex=r'.+:(?P<line>\d+):(?P<column>\d+):'
                     r'(?P<message>.+)')
class GhcModBear:
    """
    Syntax checking with ``ghc`` for Haskell files.

    See <https://hackage.haskell.org/package/ghc-mod> for more information!
    """

    LANGUAGES = {'Haskell'}
    REQUIREMENTS = {CabalRequirement(package='ghc-mod', version='5.6.0')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/98873'
    CAN_DETECT = {'Syntax'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        # -b '. ' is the argument given for ghc-mod for seperation of messages
        return '-b', '. ', 'check', filename



from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.NpmRequirement import NpmRequirement


@linter(executable='postcss',
        output_format='corrected',
        result_message='Add vendor prefixes to CSS rules.',
        prerequisite_check_command=('postcss', '--use', 'autoprefixer'),
        prerequisite_check_fail_message='Autoprefixer is not installed.')
class CSSAutoPrefixBear:
    """
    This bear adds vendor prefixes to CSS rules using ``autoprefixer`` utility.
    """
    LANGUAGES = {'CSS'}
    REQUIREMENTS = {NpmRequirement('postcss-cli', '2'),
                    NpmRequirement('autoprefixer', '6')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    ASCIINEMA_URL = 'https://asciinema.org/a/40093'
    CAN_FIX = {'Syntax', 'Formatting'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--use', 'autoprefixer', filename



import json

from coalib.bearlib import deprecate_settings
from coalib.bearlib.abstractions.Linter import linter
from dependency_management.requirements.GemRequirement import GemRequirement
from coalib.results.Result import Result
from coalib.results.SourceRange import SourceRange
from coala_utils.param_conversion import negate


@linter(executable='reek', use_stdin=True)
class RubySmellBear:
    """
    Detect code smells in Ruby source code.

    For more information about the detected smells, see
    <https://github.com/troessner/reek/blob/master/docs/Code-Smells.md>.
    """

    LANGUAGES = {'Ruby'}
    REQUIREMENTS = {GemRequirement('reek')}
    AUTHORS = {'The coala developers'}
    AUTHORS_EMAILS = {'coala-devel@googlegroups.com'}
    LICENSE = 'AGPL-3.0'
    CAN_DETECT = {'Smell'}

    @staticmethod
    def create_arguments(filename, file, config_file):
        return '--format', 'json', '-c', config_file

    def process_output(self, output, filename, file):
        output = json.loads(output) if output else ()
        for issue in output:
            sourceranges = []
            for line in issue['lines']:
                sourceranges.append(SourceRange.from_values(
                    file=filename, start_line=line))

            if 'name' in issue:
                message = "'{}' (in '{}') {}.".format(
                    issue['name'], issue['context'], issue['message'])
            else:
                message = "'{}' {}".format(issue['context'], issue['message'])

            yield Result(
                origin='{} ({})'.format(self.__class__.__name__,
                                        issue['smell_type']),
                message=message,
                affected_code=sourceranges,
                additional_info='More information is available at {}'
                                '.'.format(issue['wiki_link']))

    @deprecate_settings(allow_duplicate_method=(
                            'duplicate_method_call', negate),
                        allow_data_clump=('data_clump', negate),
                        allow_control_parameters=('control_parameter', negate),
                        allow_class_variables=('class_variable', negate),
                        allow_boolean_parameter_in_functions=(
                            'boolean_parameter', negate),
                        allow_setter_in_classes=('attribute', negate),
                        allow_unused_private_methods=(
                            'unused_private_method', negate),
                        allow_unused_variables=('unused_params', negate))
    def generate_config(self,
                        allow_setter_in_classes: bool=False,
                        allow_boolean_parameter_in_functions: bool=False,
                        allow_class_variables: bool=False,
                        allow_control_parameters: bool=False,
                        allow_data_clump: bool=False,
                        allow_duplicate_method: bool=False,
                        feature_envy: bool=True,
                        missing_module_description: bool=True,
                        long_param_list: bool=True,
                        long_yield_list: bool=True,
                        module_initialize: bool=True,
                        nested_iterators: bool=True,
                        nil_check: bool=True,
                        prima_donna_method: bool=True,
                        repeated_conditional: bool=True,
                        too_many_instance_variables: bool=True,
                        too_many_methods: bool=True,
                        too_long_method: bool=True,
                        bad_method_name: bool=True,
                        bad_module_name: bool=True,
                        bad_param_name: bool=True,
                        bad_var_name: bool=True,
                        allow_unused_variables: bool=False,
                        allow_unused_private_methods: bool=True,
                        utility_function: bool=True):
        """
        :param allow_setter_in_classes:
            Allows setter in classes.
        :param allow_boolean_parameter_in_functions:
            Allows boolean parameter in functions (control coupling).
        :param allow_class_variables:
            Allows class variables.
        :param allow_control_parameters:
            Allows parameters that control function behaviour (control
            coupling).
        :param allow_data_clump:
            Does not warn when the same two or three items frequently appear
            together in function/class parameter list.
        :param allow_duplicate_method:
            Allows having two fragments of code that look nearly identical, or
            two fragments of code that have nearly identical effects at some
            conceptual level.
        :param feature_envy:
            Occurs when a code fragment references another object more often
            than it references itself, or when several clients do the same
            series of manipulations on a particular type of object.
        :param missing_module_description:
            Warns if a module description is missing.
        :param long_param_list:
            Warns about too many parameters of functions.
        :param long_yield_list:
            Warns when a method yields a lot of arguments to the block it gets
            passed.
        :param module_initialize:
            Warns about ``#initialize`` methods in modules.
        :param nested_iterators:
            Warns when a block contains another block.
        :param nil_check:
            Warns about nil checks.
        :param prima_donna_method:
            Warns about methods whose names end with an exclamation mark.
        :param repeated_conditional:
            Warns about repeated conditionals.
        :param too_many_instance_variables:
            Warns for too many instance variables.
        :param too_many_methods:
            Warns if a class has too many methods.
        :param too_long_method:
            Warns about huge methods.
        :param bad_method_name:
            Warns about method names which are not communicating the purpose
            of the method well.
        :param bad_module_name:
            Warns about module names which are not communicating the purpose
            of the module well.
        :param bad_param_name:
            Warns about parameter names which are not communicating the purpose
            of the parameter well.
        :param bad_var_name:
            Warns about variable names which are not communicating the purpose
            of the variable well.
        :param allow_unused_variables:
            Allows unused parameters though they are dead code.
        :param check_unused_private_methods:
            Warns about unused private methods, as they are dead code.
        :param utility_function:
            Allows any instance method that has no dependency on the state of
            the instance.
        """
        config = {
            'Attribute': not allow_setter_in_classes,
            'BooleanParameter': not allow_boolean_parameter_in_functions,
            'ClassVariable': not allow_class_variables,
            'ControlParameter': not allow_control_parameters,
            'DataClump': not allow_data_clump,
            'DuplicateMethodCall': not allow_duplicate_method,
            'FeatureEnvy': feature_envy,
            'IrresponsibleModule': missing_module_description,
            'LongParameterList': long_param_list,
            'LongYieldList': long_yield_list,
            'ModuleInitialize': module_initialize,
            'NestedIterators': nested_iterators,
            'NilCheck': nil_check,
            'PrimaDonnaMethod': prima_donna_method,
            'RepeatedConditional': repeated_conditional,
            'TooManyInstanceVariables': too_many_instance_variables,
            'TooManyMethods': too_many_methods,
            'TooManyStatements': too_long_method,
            'UncommunicativeMethodName': bad_method_name,
            'UncommunicativeModuleName': bad_module_name,
            'UncommunicativeParameterName': bad_param_name,
            'UncommunicativeVariableName': bad_var_name,
            'UnusedParameters': not allow_unused_variables,
            'UnusedPrivateMethod': not allow_unused_private_methods,
            'UtilityFunction': utility_function}

        return ('---\n' +
                '\n'.join('{}:\n  enabled: {}'.format(key, str(value).lower())
                          for key, value in config.items()))
